
#include "ad++.h"

const int N = 10 ;
const int NITER = 100 ;

int wlo [] = {1, 1}, whi [] = {1, 1} ;  // ghost widths for `writeHalo'

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs2 p ;
  ON(p) {
    BlockRange x(N, p.dim(0), 1, 1), y(N, p.dim(1), 1, 1) ;

    float h = 1.0 / (N - 1) ;

    // Initialization:
    //    sol = exact solution
    //    f   = forcing term
    //    u   = approximation solution

    Array2<float> u(x, y), f(x, y), sol(x, y) ;

    const float fac = 4.0 - 2.0 * (cos(h) + cos(3 * h)) ;

    Index i(x), j(y) ;
    OVERALL(i) {
      OVERALL(j) {
	sol(i, j) = sin(3 * (h * i) + h * j) ;
        if(i == 0 || i == N - 1 || j == 0 || j == N - 1)
	  u(i, j) = sol(i, j) ;
        else {
	  u(i, j) = 0.0 ;
	  f(i, j) = fac * sol(i, j) ;
        }
      } ALLOVER(j) ;
    } ALLOVER(i) ;

    gprintf("initial u:\n") ;
    aprintf("%7.3f%*N", u, N) ;

    // Main loop

    Index2 i1(x, N - 2, 1) ;

    for(int iter = 0 ; iter < NITER ; iter++) {
      for(int parity = 0 ; parity < 2 ; parity++) {

	writeHalo(u, wlo, whi) ;

	OVERALL(i1) {
          const int base  = 1 + (1 + i1 + parity) % 2,
                    count = (N - 2 - base) / 2 + 1 ;
	  Index3 j1(y, count, base, 2) ;

	  OVERALL(j1) {
	    u(i1, j1) = 0.25 * (f(i1, j1) +
                                u(i1 + 1, j1) + u(i1 - 1, j1) +
				u(i1, j1 + 1) + u(i1, j1 - 1)) ;
	  } ALLOVER(j1) ;
	} ALLOVER(i1) ;
      }

      gprintf("iteration = %d\n", iter) ;
      aprintf("%7.3f%*N", u, N) ;
    }

    gprintf("exact solution:\n") ;
    aprintf("%7.3f%*N", sol, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

