/*
 * File         : Section4.java
 * Author       : Guansong Zhang
 * Created      : Tue Nov  4 16:44:58 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1998/02/10 23:12:22 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

package spmd;

import spmd.*;

public class Section4 extends Section {
  
  Section4 (int type) {
    super(4, type); 
    
    this.range=new Range[4];
    this.stride=new Stride[4];
  }

  public int element(Location loc0, Location loc1, Location loc2, Location loc3) {
    return(this.offset+stride[0].offset(loc0)+stride[1].offset(loc1)+stride[2].offset(loc2)+stride[3].offset(loc3));    
  };

  public Section0 subsec(Location loc0, Location loc1, Location loc2, Location loc3) {

    Section0 sec=new Section0(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc1);
    sec.group=sec.group.divide(loc2);
    sec.group=sec.group.divide(loc3);
    
    DAD dad = new DAD (sec.group);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[1].offset(loc1)+this.stride[2].offset(loc2)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section1 subsec(Range rng0, Location loc1, Location loc2, Location loc3) {

    Section1 sec=new Section1(this.type);
    
    sec.group=this.group.divide(loc1);
    sec.group=sec.group.divide(loc2);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng0;
    sec.stride[0]=this.stride[0];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[1].offset(loc1)+this.stride[2].offset(loc2)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section1 subsec(Location loc0, Range rng1, Location loc2, Location loc3) {

    Section1 sec=new Section1(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc2);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng1;
    sec.stride[0]=this.stride[1];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[2].offset(loc2)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section1 subsec(Location loc0, Location loc1, Range rng2, Location loc3) {

    Section1 sec=new Section1(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc1);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng2;
    sec.stride[0]=this.stride[2];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[1].offset(loc1)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section1 subsec(Location loc0, Location loc1, Location loc2, Range rng3) {

    Section1 sec=new Section1(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc1);
    sec.group=sec.group.divide(loc2);

    sec.range[0]=rng3;
    sec.stride[0]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[1].offset(loc1)+this.stride[2].offset(loc2);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section2 subsec(Range rng0, Range rng1, Location loc2, Location loc3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc2);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng0;
    sec.range[1]=rng1;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[1];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[2].offset(loc2)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
  }

  public Section2 subsec(Range rng0, Location loc1, Range rng2, Location loc3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc1);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng0;
    sec.range[1]=rng2;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[2];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[1].offset(loc1)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
  }

  public Section2 subsec(Range rng0, Location loc1, Location loc2, Range rng3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc1);
    sec.group=sec.group.divide(loc2);

    sec.range[0]=rng0;
    sec.range[1]=rng3;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[1].offset(loc1)+this.stride[2].offset(loc2);

    sec.reference=this.reference;

    return sec;
  }

  public Section2 subsec(Location loc0, Range rng1, Range rng2, Location loc3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc3);

    sec.range[0]=rng1;
    sec.range[1]=rng2;
    sec.stride[0]=this.stride[1];
    sec.stride[1]=this.stride[2];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
  }

  public Section2 subsec(Location loc0, Range rng1, Location loc2, Range rng3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc2);

    sec.range[0]=rng1;
    sec.range[1]=rng3;
    sec.stride[0]=this.stride[1];
    sec.stride[1]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[2].offset(loc2);

    sec.reference=this.reference;

    return sec;
  }

  public Section2 subsec(Location loc0, Location loc1, Range rng2, Range rng3) {

    Section2 sec=new Section2(this.type);
    
    sec.group=this.group.divide(loc0);
    sec.group=sec.group.divide(loc1);

    sec.range[0]=rng2;
    sec.range[1]=rng3;
    sec.stride[0]=this.stride[2];
    sec.stride[1]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0)+this.stride[1].offset(loc1);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section3 subsec(Range rng0, Range rng1, Range rng2, Location loc3) {

    Section3 sec=new Section3(this.type);
    
    sec.group=this.group.divide(loc3);

    sec.range[0]=rng0;
    sec.range[1]=rng1;
    sec.range[2]=rng2;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[1];
    sec.stride[2]=this.stride[2];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[3].offset(loc3);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section3 subsec(Range rng0, Range rng1, Location loc2, Range rng3) {

    Section3 sec=new Section3(this.type);
    
    sec.group=this.group.divide(loc2);

    sec.range[0]=rng0;
    sec.range[1]=rng1;
    sec.range[2]=rng3;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[1];
    sec.stride[2]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[2].offset(loc2);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section3 subsec(Range rng0, Location loc1, Range rng2, Range rng3) {

    Section3 sec=new Section3(this.type);
    
    sec.group=this.group.divide(loc1);

    sec.range[0]=rng0;
    sec.range[1]=rng2;
    sec.range[2]=rng3;
    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[2];
    sec.stride[2]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[1].offset(loc1);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section3 subsec(Location loc0, Range rng1, Range rng2, Range rng3) {

    Section3 sec=new Section3(this.type);
    
    sec.group=this.group.divide(loc0);

    sec.range[0]=rng1;
    sec.range[1]=rng2;
    sec.range[2]=rng3;
    sec.stride[0]=this.stride[1];
    sec.stride[1]=this.stride[2];
    sec.stride[2]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset+this.stride[0].offset(loc0);

    sec.reference=this.reference;

    return sec;
    
  }

  public Section4 subsec(Range rng0, Range rng1, Range rng2, Range rng3) {

    Section4 sec=new Section4(this.type);
    
    sec.group=this.group;

    sec.range[0]=rng0;
    sec.range[1]=rng1;
    sec.range[2]=rng2;
    sec.range[3]=rng3;

    sec.stride[0]=this.stride[0];
    sec.stride[1]=this.stride[1];
    sec.stride[2]=this.stride[2];
    sec.stride[2]=this.stride[3];
    
    DAD dad = new DAD (sec.group, sec.range, sec.stride);
    sec.dad=dad;

    sec.offset=this.offset;

    sec.reference=this.reference;

    return sec;
    
  }

}
