/*
 * File         : spmd_DAD.C
 * Author       : Guansong Zhang
 * Created      : Fri Oct 24 20:52:55 1997
 * Revision     : $Revision: 1.3 $
 * Updated      : $Date: 1998/09/29 19:48:08 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "spmd_DAD.h"

#include "adlib.h"

Group * spmd_Group_ResolvePtr(JNIEnv *, jobject);
Range * spmd_Range_ResolvePtr(JNIEnv *, jobject);
Stride * spmd_Stride_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;

#include "jctable.h"
jctable<DAD> spmd_DAD_Table;

/*ResolvePtr
 */
DAD * spmd_DAD_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(spmd_DAD_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     spmd_DAD
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_DAD_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("spmd/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<DAD> *tempspmd_DAD_Table=new jctable<DAD>;
  spmd_DAD_Table=*tempspmd_DAD_Table;
  delete tempspmd_DAD_Table;
};

/*
 * Class:     spmd_DAD
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_DAD_clean
  (JNIEnv *, jclass);

/*
 * Class:     spmd_DAD
 * Method:    constructor
 * Signature: (Lspmd/Group;)V
 */

JNIEXPORT void JNICALL Java_spmd_DAD_constructor__Lspmd_Group_2
  (JNIEnv * env, jobject jthis, jobject jGroup)
{
  Group * groupPtr = spmd_Group_ResolvePtr(env, jGroup); 

  // this is for rank-0 array
  DAD *cppDAD=new DAD(0, *groupPtr);
  long pos = spmd_DAD_Table.NewHandle(cppDAD);

  if (spmd_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //spmd_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     spmd_DAD
 * Method:    constructor
 * Signature: (Lspmd/Group;[Lspmd/Range;[Lspmd/Stride;)V
 */
JNIEXPORT void JNICALL Java_spmd_DAD_constructor__Lspmd_Group_2_3Lspmd_Range_2_3Lspmd_Stride_2
  (JNIEnv * env, jobject jthis, jobject jGroup, jobjectArray jRangeArray, jobjectArray jStrideArray)
{
  Group * groupPtr = spmd_Group_ResolvePtr(env, jGroup); 

  // the rank of Range and Stride will be the same.

  int rank=(env->GetArrayLength(jRangeArray));

  // Range and Stride should have default constructor

  Range * rangeLst = new Range [rank];
  Stride * strideLst = new Stride [rank];
  int * sizeLst = new int [rank];

  jclass Stride_class=env->FindClass("spmd/Stride");
  jfieldID sizeID=env->GetFieldID(Stride_class, "size", "I");;

  for (int i = 0; i<rank; i++) {
    jobject jRange=env->GetObjectArrayElement(jRangeArray, i);
    *(rangeLst+i)=(*spmd_Range_ResolvePtr(env, jRange));

    jobject jStride=env->GetObjectArrayElement(jStrideArray, i);
    *(strideLst+i)=(*spmd_Stride_ResolvePtr(env, jStride));

    // assign sizeList
    *(sizeLst+i)=(strideLst+i)->itemLen;
  }

  DAD *cppDAD=new DAD(rank, *groupPtr, rangeLst, strideLst);
  long pos= spmd_DAD_Table.NewHandle(cppDAD);

  //printf("DAD: %d\n",pos);

  if (spmd_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }


  //  AdlibPrintf("group address is %x\n", groupPtr); // debug

  //is this a problem?
  //delete [] rangeLst;
  //delete [] strideLst;

  //spmd_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

};

/*
 * Class:     spmd_DAD
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_DAD_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=spmd_DAD_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    // first delete two arrays for the object

    delete spmd_DAD_Table.ObjPtr[thecppobj];
    spmd_DAD_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     spmd_DAD
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_spmd_DAD_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  DAD * dad=spmd_DAD_ResolvePtr(env, jthis);

  AdlibPrintf("DAD %s, Handle %d, Address %x ", cname, pos, dad);

  printf("Rank %d, (Group %x), RangeList %x, StrideList %x\n", 
	 dad->rank, &dad->group, dad->ranges, dad->strides);

  delete [] cname;

};
