/*
 * File         : spmd_Group.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 20 09:45:42 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1998/02/10 23:12:24 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "spmd_Group.h"
#include "adlib.h"
#include "apgStack.h"

extern jfieldID hCPPObjID;

Procs * spmd_Procs_ResolvePtr(JNIEnv *, jobject);
Location * spmd_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Group> spmd_Group_Table;

/*ResolvePtr
 */
Group * spmd_Group_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(spmd_Group_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     spmd_Group
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Group_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("spmd/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Group> *tempspmd_Group_Table=new jctable<Group>;
  spmd_Group_Table=*tempspmd_Group_Table;
  delete tempspmd_Group_Table;
};

/*
 * Class:     spmd_Group
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Group_clean
  (JNIEnv *, jclass)
{
//  delete spmd_Group_ObjPtr;
//  delete spmd_Group_ObjRef;
};

/*
 * Class:     spmd_Group
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Group_constructor
  (JNIEnv *env, jobject jthis)
{
  Group *cppGroup=new Group();
  long pos=spmd_Group_Table.NewHandle(cppGroup);

  if (spmd_Group_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  // spmd_Group_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     spmd_Group
 * Method:    constructor
 * Signature: (Lspmd/Procs;)V
 */
/*
JNIEXPORT void JNICALL Java_spmd_Group_constructor__Lspmd_Procs_2
  (JNIEnv *env, jobject jthis,jobject p)
{
  Group *cppGroup=new Group(*spmd_Procs_ResolvePtr(env,p));
  long pos=spmd_Group_Table.Newhandle(cppGroup);

  if (spmd_Group_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

 // spmd_Group_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);
};
*/

/*
 * Class:     spmd_Group
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Group_finalize
  (JNIEnv * env, jobject jthis)
{
  int thecppobj=env->GetIntField(jthis, hCPPObjID);
  int rc=spmd_Group_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete spmd_Group_Table.ObjPtr[thecppobj];
    spmd_Group_Table.ObjPtr[thecppobj]=NULL;
  }

};

/*
 * Class:     spmd_Group
 * Method:    on
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_spmd_Group_on
  (JNIEnv * env, jobject jthis)
{
  Group * cppGroup=spmd_Group_ResolvePtr(env, jthis); 

  apgStack.push(&apg);
  apg = *cppGroup;

  int temp = cppGroup->member();
  //printf("testing on: %d\n", temp); // debug
  
  if (temp)
    return (JNI_TRUE);
  else
    return (JNI_FALSE);

  //  return(JNI_TRUE);
};

/*
 * Class:     spmd_Group
 * Method:    no
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Group_no
  (JNIEnv * env, jobject jthis)
{  
  apgStack.pop(&apg);
};


/*
 * Class:     spmd_Group
 * Method:    id
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_Group_id
  (JNIEnv * env, jobject jthis)
{
  return (spmd_Group_ResolvePtr(env, jthis))->id();
}

/*
 * Class:     spmd_Group
 * Method:    divide
 * Signature: (Lspmd/Location;)Lspmd/Group;
 */
JNIEXPORT jobject JNICALL Java_spmd_Group_divide
  (JNIEnv *env, jobject jthis, jobject i)
{
  //Group *newCPPgroupPtr=new Group();

  Group CPPgroup = (*spmd_Group_ResolvePtr(env, jthis))/(*spmd_Location_ResolvePtr(env,i));

  //*newCPPgroupPtr=CPPgroup;

  //long pos=spmd_Group_NewHandle(newCPPgroupPtr);

  jclass ccGroup=NULL;
  ccGroup=env->FindClass("spmd/Group");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccGroup, "<init>", "()V");
  jobject jNode=env->NewObject(ccGroup, handleConstructorID);

  Group *newCPPgroupPtr=spmd_Group_ResolvePtr(env,jNode);

  *newCPPgroupPtr=CPPgroup;
  return(jNode);
}

/*
 * Class:     spmd_Group
 * Method:    restrict
 * Signature: (Lspmd/Location;)V
 */
JNIEXPORT void JNICALL Java_spmd_Group_restrict
  (JNIEnv * env , jobject jthis, jobject i)
{
  (*spmd_Group_ResolvePtr(env, jthis))/=(*spmd_Location_ResolvePtr(env,i));
};

/*
 * Class:     spmd_Group
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_spmd_Group_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Group %s, Handle %d, Address %x\n", cname, pos, spmd_Group_ResolvePtr(env, jthis));

  spmd_Group_ResolvePtr(env, jthis)->show();

  delete [] cname;

};

