/*
 * File         : spmd_LocBlocksIndex.C
 * Author       : Guansong Zhang
 * Created      : Wed Jan  7 14:42:39 1998
 * Revision     : $Revision: 1.4 $
 * Updated      : $Date: 1998/02/20 17:00:05 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include "spmd_LocBlocksIndex.h"

#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;

jfieldID countID=0;
jfieldID loopingID=0;

#include "jctable.h"

extern jctable<Location> spmd_Location_Table;

Location * spmd_Location_ResolvePtr(JNIEnv *, jobject);
Range * spmd_Range_ResolvePtr(JNIEnv *, jobject);

/*ResolvePtr
 */
LocBlocksIndex * spmd_LocBlocksIndex_ResolvePtr(JNIEnv *env, jobject jthis) {
  return((LocBlocksIndex *)spmd_Location_ResolvePtr(env, jthis));
}

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_LocBlocksIndex_init
  (JNIEnv * env , jclass jthis)
{
  jclass LocBlocksIndex_class=env->FindClass("spmd/LocBlocksIndex");
  countID=env->GetFieldID(LocBlocksIndex_class, "count", "I");
  loopingID=env->GetFieldID(LocBlocksIndex_class, "looping", "I");
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    constructor
 * Signature: (Lspmd/Range;)V
 */
JNIEXPORT void JNICALL Java_spmd_LocBlocksIndex_constructor
  (JNIEnv * env, jobject jthis, jobject jRange)
{
  LocBlocksIndex * cppLocBlocksIndex = new LocBlocksIndex(*spmd_Range_ResolvePtr(env, jRange));

  int pos = env->GetIntField(jthis, hCPPObjID);

  delete (spmd_Location_Table.ObjPtr[pos]);
  
  spmd_Location_Table.ObjPtr[pos]=cppLocBlocksIndex;

  if (spmd_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }
  
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    beginLocBlk
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_LocBlocksIndex_beginLocBlk
  (JNIEnv * env, jobject jthis)
{
  spmd_LocBlocksIndex_ResolvePtr(env, jthis)->beginLocBlk();
};


/*
 * Class:     spmd_LocBlocksIndex
 * Method:    test
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_spmd_LocBlocksIndex_test
  (JNIEnv * env, jobject jthis)
{
  if (spmd_LocBlocksIndex_ResolvePtr(env, jthis)->test())
    return (JNI_TRUE);
  else
    return (JNI_FALSE);
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    nextLocBlk
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_LocBlocksIndex_nextLocBlk
  (JNIEnv * env, jobject jthis)
{
  spmd_LocBlocksIndex_ResolvePtr(env, jthis)->nextLocBlk();
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    count
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_LocBlocksIndex_count
  (JNIEnv * env, jobject jthis)
{
  return (spmd_LocBlocksIndex_ResolvePtr(env, jthis)->count);
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    glb_bas
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_LocBlocksIndex_glb_1bas
  (JNIEnv * env, jobject jthis)
{
  return (spmd_LocBlocksIndex_ResolvePtr(env, jthis)->glb_bas);
};

/*
 * Class:     spmd_LocBlocksIndex
 * Method:    glb_stp
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_LocBlocksIndex_glb_1stp
  (JNIEnv * env, jobject jthis)
{
  return (spmd_LocBlocksIndex_ResolvePtr(env, jthis)->glb_stp);
};
