/*
 * File         : spmd_Range.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:06:49 1997
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1998/09/29 19:48:10 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "spmd_Range.h"
#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;
#include "jctable.h"
extern jctable<Dimension> spmd_Dimension_Table;

Dimension * spmd_Dimension_ResolvePtr(JNIEnv *, jobject);
Stride *spmd_Stride_ResolvePtr(JNIEnv *, jobject);
Location * spmd_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Range> spmd_Range_Table;

/*ResolvePtr
 */
Range * spmd_Range_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(spmd_Range_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     spmd_Range
 * Method:    idx
 * Signature: (Lspmd/Location;)I
 */
JNIEXPORT jint JNICALL Java_spmd_Range_idx
  (JNIEnv * env, jobject jthis, jobject jlocation)
{
  return (spmd_Range_ResolvePtr(env, jthis)->idx(*spmd_Location_ResolvePtr(env, jlocation)));
};

/*
 * Class:     spmd_Range
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Range_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("spmd/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Range> *tempspmd_Range_Table=new jctable<Range>;
  spmd_Range_Table=*tempspmd_Range_Table;
  delete tempspmd_Range_Table;
};

/*
 * Class:     spmd_Range
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Range_clean
  (JNIEnv *, jclass);

/*
 * Class:     spmd_Range
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Range_finalize
  (JNIEnv * env, jobject jthis)
{
  int thecppobj=env->GetIntField(jthis, hCPPObjID);
  int rc=spmd_Range_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete spmd_Range_Table.ObjPtr[thecppobj];
    spmd_Range_Table.ObjPtr[thecppobj]=NULL;
  }
}

/*
 * Class:     spmd_Range
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_Range_constructor__
  (JNIEnv *env, jobject jthis)
{
  Range *cppRange= new Range();
  long pos=spmd_Range_Table.NewHandle(cppRange);

  if (spmd_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //spmd_Range_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

}

/*
 * Class:     spmd_Range
 * Method:    constructor
 * Signature: (I)V
 */
/*
JNIEXPORT void JNICALL Java_spmd_Range_constructor__I
  (JNIEnv * env, jobject jthis, jint extend)
{
  Range *cppRange= new Range(extend);
  long pos=spmd_Range_Table.NewHandle(cppRange);

  if (spmd_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //spmd_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);
};

*/

/*
 * Class:     spmd_Range
 * Method:    constructor
 * Signature: (Lspmd/Dimension;)V
 */
/*
JNIEXPORT void JNICALL Java_spmd_Range_constructor__Lspmd_Dimension_2
  (JNIEnv *env, jobject jthis, jobject dim) 
{
  //  (new Range(*spmd_Dimension_ResolvePtr(env,dim)))->show();
  Range *cppRange= new Range(*spmd_Dimension_ResolvePtr(env,dim));
  long pos=spmd_Range_Table.NewHandle(cppRange);
  
  //spmd_Dimension_ResolvePtr(env,dim)->show();
  //spmd_Range_ObjPtr[pos]->show(); //debug;

  if (spmd_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //spmd_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);

}
*/

/*
 * Class:     spmd_Range
 * Method:    dim
 * Signature: (I)Lspmd/Dimension;
 */
JNIEXPORT jobject JNICALL Java_spmd_Range_dim
  (JNIEnv *env, jobject jthis)
{

  //Dimension *newCPPdimPtr=new Dimension();

  Dimension CPPdim = spmd_Range_ResolvePtr(env, jthis)->dim();
  
  //*newCPPdimPtr=CPPdim;

  //long pos=spmd_Dimension_NewHandle(newCPPdimPtr);

  //printf("dim pos %ld, address %x\n", pos, spmd_Dimension_ObjPtr[pos]);
  //printf("is this dim? crd %d\n", spmd_Dimension_ObjPtr[pos]->crd());

  jclass ccDimension=NULL;
  ccDimension=env->FindClass("spmd/Dimension");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccDimension, "<init>", "()V");
  jobject jNode=env->NewObject(ccDimension, handleConstructorID);

  Dimension * newCPPdimPtr=spmd_Dimension_ResolvePtr(env, jNode);

  *newCPPdimPtr=CPPdim;

  return(jNode);
};

/*
 * Class:     spmd_Range
 * Method:    subrng
 * Signature: (III)Lspmd/Range;
 */
JNIEXPORT jobject JNICALL Java_spmd_Range_subrng
  (JNIEnv *env, jobject jthis, jint extent, jint base, jint stride)
{
  Range CPPrange = spmd_Range_ResolvePtr(env, jthis)->subrng(extent,base,stride);

  jclass ccRange=NULL;
  ccRange=env->FindClass("spmd/Range");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccRange, "<init>", "()V");
  jobject jNode=env->NewObject(ccRange, handleConstructorID);

  Range * newCPPrangePtr=spmd_Range_ResolvePtr(env, jNode);

  *newCPPrangePtr=CPPrange;

  return(jNode);
  
}

/*
 * Class:     spmd_Range
 * Method:    triplet
 * Signature: (III)Lspmd/Range;
 */
JNIEXPORT jobject JNICALL Java_spmd_Range_triplet
  (JNIEnv * env, jobject jthis, jint base, jint top, jint stride)
{
  Range CPPrange = spmd_Range_ResolvePtr(env, jthis)->triplet(base, top, stride);

  jclass ccRange=NULL;
  ccRange=env->FindClass("spmd/Range");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccRange, "<init>", "()V");
  jobject jNode=env->NewObject(ccRange, handleConstructorID);

  Range * newCPPrangePtr=spmd_Range_ResolvePtr(env, jNode);

  *newCPPrangePtr=CPPrange;

  return(jNode);
};

/*
 * Class:     spmd_Range
 * Method:    size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_Range_size
  (JNIEnv * env, jobject jthis)
{
  int temp = spmd_Range_ResolvePtr(env, jthis)->size();
  //printf("testing on: %d\n", temp);
  //cout<<"testing on: " << temp <<"\n";
  return(temp);
};

/*
 * Class:     spmd_Range
 * Method:    lev
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_Range_lev
  (JNIEnv * env, jobject jthis)
{
	return(spmd_Range_ResolvePtr(env, jthis)->lev());
}

/*
 * Class:     spmd_Range
 * Method:    bas
 * Signature: ()I
 */
JNIEXPORT  jint JNICALL Java_spmd_Range_bas
  (JNIEnv * env, jobject jthis)
{
  return(spmd_Range_ResolvePtr(env, jthis)->bas());  
}

/*
 * Class:     spmd_Range
 * Method:    str
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_Range_str
  (JNIEnv *env, jobject jthis)
{
  return(spmd_Range_ResolvePtr(env, jthis)->str());
}

/*
 * Class:     spmd_Range
 * Method:    ker_glb
 * Signature: (I[I)I
 */
/*
JNIEXPORT jint JNICALL Java_spmd_Range_ker_1glb
  (JNIEnv *env, jobject jthis, jint glb, jintArray sub)
{
  int size=(env->GetArrayLength(sub));
  int * cppsub=new int [size];

  jboolean isCopy;
  jint * jsub;

  jsub=env->GetIntArrayElements(sub, &isCopy);

  for (int i=0; i<size; i++)
    *(cppsub+i)=*(jsub+i);

  env->ReleaseIntArrayElements(sub, jsub, 0);

  return(spmd_Range_ResolvePtr(env, jthis)->ker_glb(glb,cppsub));
}
*/

/*
 * Class:     spmd_Range
 * Method:    width
 * Signature: ()I
 */
// JNIEXPORT jint JNICALL Java_spmd_Range_width
//   (JNIEnv *env, jobject jthis)
// {
//     return(spmd_Range_ResolvePtr(env, jthis)->width());
// }

/*
 * Class:     spmd_Range
 * Method:    volume
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_spmd_Range_volume
  (JNIEnv *env, jobject jthis)
{
  return(spmd_Range_ResolvePtr(env, jthis)->volume());
}
//inline Stride Range :: stride(int* size) const {
//  const int itemLen = *size ;
//  *size *= volume() ;
//  return Stride(*this, itemLen) ;
//}

/*
 * Class:     spmd_Range
 * Method:    stride
 * Signature: (I)Lspmd/Stride;
 */
JNIEXPORT jobject JNICALL Java_spmd_Range_stride
  (JNIEnv *env, jobject jthis, jint size)
{
  //Stride *newCPPstridePtr=new Stride();

  int ccsize=size;
  Range parent=*spmd_Range_ResolvePtr(env, jthis);

  Stride CPPstride(parent, ccsize);
  ccsize*=parent.volume();
 
  //printf("range address %x, stride size %d\n",spmd_Range_ResolvePtr(env, jthis), ccsize); //debug

  //*newCPPstridePtr=CPPstride;

  //long pos=spmd_Stride_NewHandle(newCPPstridePtr);

  jclass ccStride=NULL;
  ccStride=env->FindClass("spmd/Stride");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccStride, "<init>", "()V");
  jobject jNode=env->NewObject(ccStride, handleConstructorID);

  jfieldID jsize=env->GetFieldID(ccStride,"size","I");
  env->SetIntField(jNode,jsize,ccsize);

  Stride *newCPPstridePtr=spmd_Stride_ResolvePtr(env,jNode);
  *newCPPstridePtr=CPPstride;

  return(jNode);
}


/*
 * Class:     spmd_Range
 * Method:    location
 * Signature: (I)Lspmd/Location;
 */
JNIEXPORT jobject JNICALL Java_spmd_Range_location
   (JNIEnv * env, jobject jthis, jint i)
{
 
  Location CPPlocation = (*spmd_Range_ResolvePtr(env, jthis))(i);
  
  //*newCPPstridePtr=CPPstride;
  
  //long pos=spmd_Stride_NewHandle(newCPPstridePtr);
  
  jclass ccLocation=NULL;
  ccLocation=env->FindClass("spmd/Location");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccLocation, "<init>", "()V");
  jobject jNode=env->NewObject(ccLocation, handleConstructorID);
  
  Location *newCPPlocationPtr=spmd_Location_ResolvePtr(env,jNode);
  
  *newCPPlocationPtr=CPPlocation;
  
  return(jNode);
   
};

/*
 * Class:     spmd_Range
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_spmd_Range_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Range %s, Handle %d, Address %x\n", cname, pos, spmd_Range_ResolvePtr(env, jthis));

  spmd_Range_ResolvePtr(env, jthis)->show();
  
  delete [] cname;

};
