/*
 * File         : spmd_WriteHalo.C
 * Author       : Guansong Zhang
 * Created      : Thu Jan 22 12:02:26 1998
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1998/01/30 17:53:36 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include "spmd_WriteHalo.h"
#include "adlib.h"

DAD * spmd_DAD_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;
#include "jctable.h"

jctable<WriteHalo> spmd_WriteHalo_Table;

/*ResolvePtr
 */
WriteHalo * spmd_WriteHalo_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(spmd_WriteHalo_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     spmd_WriteHalo
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("spmd/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");  
  jctable<WriteHalo> *tempspmd_WriteHalo_Table=new jctable<WriteHalo>;
  spmd_WriteHalo_Table=*tempspmd_WriteHalo_Table;
  delete tempspmd_WriteHalo_Table;
};

/*
 * Class:     spmd_WriteHalo
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_clean
  (JNIEnv *, jclass);

/*
 * Class:     spmd_WriteHalo
 * Method:    constructor
 * Signature: (Lspmd/DAD;I[I[I[I)V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_constructor__Lspmd_DAD_2I_3I_3I_3I
  (JNIEnv * env, jobject jthis, jobject sDAD, jint len, jintArray jwlo, jintArray jwhi, jintArray jmode)
{
  DAD * sDADPtr = spmd_DAD_ResolvePtr(env, sDAD); 

  jint * wlo=env->GetIntArrayElements(jwlo, 0);
  jint * whi=env->GetIntArrayElements(jwhi, 0);
  jint * mode=env->GetIntArrayElements(jmode, 0);

  // (jint *) != (int *)
  WriteHalo *cppWriteHalo=new WriteHalo(sDADPtr, len, (int *)wlo, (int *)whi, (Mode *)mode);

  env->ReleaseIntArrayElements(jwlo, wlo, JNI_ABORT);
  env->ReleaseIntArrayElements(jwhi, wlo, JNI_ABORT);
  env->ReleaseIntArrayElements(jmode, mode, JNI_ABORT);

  long pos=spmd_WriteHalo_Table.NewHandle(cppWriteHalo);

  if (spmd_WriteHalo_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  env->SetIntField(jthis, hCPPObjID, pos);  

};

/*
 * Class:     spmd_WriteHalo
 * Method:    constructor
 * Signature: (Lspmd/DAD;I[I[I)V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_constructor__Lspmd_DAD_2I_3I_3I
  (JNIEnv * env, jobject jthis, jobject sDAD, jint len, jintArray jwlo, jintArray jwhi)
{
  DAD * sDADPtr = spmd_DAD_ResolvePtr(env, sDAD); 

  jint * wlo=env->GetIntArrayElements(jwlo, 0);
  jint * whi=env->GetIntArrayElements(jwhi, 0);

  // (jint *) != (int *)
  WriteHalo *cppWriteHalo=new WriteHalo(sDADPtr, len, (int *)wlo, (int *)whi);

  env->ReleaseIntArrayElements(jwlo, wlo, JNI_ABORT);
  env->ReleaseIntArrayElements(jwhi, wlo, JNI_ABORT);

  long pos=spmd_WriteHalo_Table.NewHandle(cppWriteHalo);

  if (spmd_WriteHalo_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  env->SetIntField(jthis, hCPPObjID, pos);  
};


/*
 * Class:     spmd_WriteHalo
 * Method:    execute
 * Signature: ([IJ)V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_execute___3IJ
  (JNIEnv * env, jobject jthis, jintArray src, jlong sOffset)
{
  jint * jSA=env->GetIntArrayElements(src, 0);

  spmd_WriteHalo_ResolvePtr(env, jthis)->execute(jSA+sOffset); 

  env->ReleaseIntArrayElements(src, jSA, 0);
};


/*
 * Class:     spmd_WriteHalo
 * Method:    execute
 * Signature: ([FJ)V
 */
JNIEXPORT void JNICALL Java_spmd_WriteHalo_execute___3FJ
  (JNIEnv * env, jobject jthis, jfloatArray src, jlong sOffset)
{
  jfloat * jSA=env->GetFloatArrayElements(src, 0);

  spmd_WriteHalo_ResolvePtr(env, jthis)->execute(jSA+sOffset); 

  env->ReleaseFloatArrayElements(src, jSA, 0);
};
