/*
 * File         : hpjava_lang_DAD.C
 * Author       : Guansong Zhang
 * Created      : Fri Oct 24 20:52:55 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/03/19 17:56:38 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "hpjava_lang_DAD.h"

#include "adlib.h"

Group * hpjava_lang_Group_ResolvePtr(JNIEnv *, jobject);
Range * hpjava_lang_Range_ResolvePtr(JNIEnv *, jobject);
Stride * hpjava_lang_Stride_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;

#include "jctable.h"
jctable<DAD> hpjava_lang_DAD_Table;

/*ResolvePtr
 */
DAD * hpjava_lang_DAD_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(hpjava_lang_DAD_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     hpjava_lang_DAD
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_DAD_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("hpjava/lang/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<DAD> *temphpjava_lang_DAD_Table=new jctable<DAD>;
  hpjava_lang_DAD_Table=*temphpjava_lang_DAD_Table;
  delete temphpjava_lang_DAD_Table;
};

/*
 * Class:     hpjava_lang_DAD
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_DAD_clean
  (JNIEnv *, jclass);

/*
 * Class:     hpjava_lang_DAD
 * Method:    constructor
 * Signature: (Lhpjava/lang/Group;)V
 */

JNIEXPORT void JNICALL Java_hpjava_lang_DAD_constructor__Lhpjava_lang_Group_2
  (JNIEnv * env, jobject jthis, jobject jGroup)
{
  Group * groupPtr = hpjava_lang_Group_ResolvePtr(env, jGroup); 

  // this is for rank-0 array
  DAD *cppDAD=new DAD(0, *groupPtr);
  long pos = hpjava_lang_DAD_Table.NewHandle(cppDAD);

  if (hpjava_lang_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_lang_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     hpjava_lang_DAD
 * Method:    constructor
 * Signature: (Lhpjava/lang/Group;[Lhpjava/lang/Range;[Lhpjava/lang/Stride;)V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_DAD_constructor__Lhpjava_lang_Group_2_3Lhpjava_lang_Range_2_3Lhpjava_lang_Stride_2
  (JNIEnv * env, jobject jthis, jobject jGroup, jobjectArray jRangeArray, jobjectArray jStrideArray)
{
  Group * groupPtr = hpjava_lang_Group_ResolvePtr(env, jGroup); 

  // the rank of Range and Stride will be the same.

  int rank=(env->GetArrayLength(jRangeArray));

  // Range and Stride should have default constructor

  Range * rangeLst = new Range [rank];
  Stride * strideLst = new Stride [rank];
  int * sizeLst = new int [rank];

  jclass Stride_class=env->FindClass("hpjava/lang/Stride");
  jfieldID sizeID=env->GetFieldID(Stride_class, "size", "I");;

  for (int i = 0; i<rank; i++) {
    jobject jRange=env->GetObjectArrayElement(jRangeArray, i);
    *(rangeLst+i)=(*hpjava_lang_Range_ResolvePtr(env, jRange));

    jobject jStride=env->GetObjectArrayElement(jStrideArray, i);
    *(strideLst+i)=(*hpjava_lang_Stride_ResolvePtr(env, jStride));

    // assign sizeList
    *(sizeLst+i)=(strideLst+i)->itemLen;
  }

  DAD *cppDAD=new DAD(rank, *groupPtr, rangeLst, strideLst);
  long pos= hpjava_lang_DAD_Table.NewHandle(cppDAD);

  //printf("DAD: %d\n",pos);

  if (hpjava_lang_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }


  //  AdlibPrintf("group address is %x\n", groupPtr); // debug

  //is this a problem?
  //delete [] rangeLst;
  //delete [] strideLst;

  //hpjava_lang_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

};

/*
 * Class:     hpjava_lang_DAD
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_DAD_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=hpjava_lang_DAD_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    // first delete two arrays for the object

    delete hpjava_lang_DAD_Table.ObjPtr[thecppobj];
    hpjava_lang_DAD_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     hpjava_lang_DAD
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_DAD_show
  (JNIEnv * env, jobject jthis, jstring message)
{
  int length=env->GetStringUTFLength(message);

  char * cmessage=new char[length+1];

  const char * jmessagePtr=env->GetStringUTFChars(message, 0);

  for (int i=0; i<length; i++)
    cmessage[i]=*(jmessagePtr+i);
  cmessage[length]='\0';

  env->ReleaseStringUTFChars(message, jmessagePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  DAD * dad=hpjava_lang_DAD_ResolvePtr(env, jthis);

  AdlibPrintf("DAD %s, Handle %d, Address %x ", cmessage, pos, dad);

  printf("Rank %d, (Group %x), RangeList %x, StrideList %x\n", 
	 dad->rank, &dad->group, dad->ranges, dad->strides);

  delete [] cmessage;

};
