/*
 * File         : hpjava_lang_Location.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:03:03 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/03/19 17:56:41 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "hpjava_lang_Location.h"

#include "adlib.h"

#include "apgStack.h"

extern jfieldID hCPPObjID;

Range * hpjava_lang_Range_ResolvePtr(JNIEnv *, jobject);
Dimension * hpjava_lang_Dimension_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Location> hpjava_lang_Location_Table;


/*ResolvePtr
 */
Location * hpjava_lang_Location_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(hpjava_lang_Location_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     hpjava_lang_Location
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("hpjava/lang/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Location> *temphpjava_lang_Location_Table=new jctable<Location>;
  hpjava_lang_Location_Table=*temphpjava_lang_Location_Table;
  delete temphpjava_lang_Location_Table;
};

/*
 * Class:     hpjava_lang_Location
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_clean
  (JNIEnv *, jclass)
{
//  delete hpjava_lang_Location_ObjPtr;
//  delete hpjava_lang_Location_ObjRef;
};

/*
 * Class:     hpjava_lang_Location
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_constructor__
  (JNIEnv * env, jobject jthis)
{
  Location *cppLocation=new Location();
  long pos=hpjava_lang_Location_Table.NewHandle(cppLocation);

  if (hpjava_lang_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //AdlibPrintf("hpjava_lang_Location_AllocLength %d\n",hpjava_lang_Location_AllocLength ); // debug 
  //AdlibPrintf("new location handle is %d\n", pos); // debug

  //hpjava_lang_Location_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     hpjava_lang_Location
 * Method:    constructor
 * Signature: (Lhpjava/lang/Range;I)V
 */
/*
JNIEXPORT void JNICALL Java_hpjava_lang_Location_constructor__Lhpjava_lang_Range_2I
  (JNIEnv *env, jobject jthis, jobject x, jint glb)
{
  Location *cppLocation=new Location(*hpjava_lang_Range_ResolvePtr(env,x),glb);
  long pos=hpjava_lang_Location_Table.NewHandle(cppLocation);

  if (hpjava_lang_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_lang_Location_IncRefCntr(pos);

  env->SetIntField(jthis, hCPPObjID, pos);
}
*/

/*
 * Class:     hpjava_lang_Location
 * Method:    ta
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_ta
  (JNIEnv *env, jobject jthis)
{
  apgStack.pop(&apg) ;
}

/*
 * Class:     hpjava_lang_Location
 * Method:    at
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_hpjava_lang_Location_at
  (JNIEnv * env, jobject jthis)
{
  apgStack.push(&apg);

  Location * location=hpjava_lang_Location_ResolvePtr(env, jthis);

  apg.restrict(location->dim);

  if (location->dim.crd()==location->crd)
    return (JNI_TRUE);
  else
    return (JNI_FALSE);

  return(JNI_TRUE);
};

/*
 * Class:     hpjava_lang_Location
 * Method:    dim
 * Signature: ()Lhpjava/lang/Dimension;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Location_dim
  (JNIEnv * env, jobject jthis)
{
  Dimension cppDim = (hpjava_lang_Location_ResolvePtr(env, jthis)->dim);

  jclass ccDimension=NULL;
  ccDimension=env->FindClass("hpjava/lang/Dimension");

  if (env->ExceptionOccurred()) return 0;

  jmethodID handleConstructorID = env->GetMethodID(ccDimension, "<init>", "()V");
  jobject jDimension=env->NewObject(ccDimension, handleConstructorID);

  Dimension * newCPPdimPtr=hpjava_lang_Dimension_ResolvePtr(env, jDimension);

  *newCPPdimPtr=cppDim;

  return(jDimension);
};

/*
 * Class:     hpjava_lang_Location
 * Method:    crd
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Location_crd
  (JNIEnv * env, jobject jthis)
{
  return(hpjava_lang_Location_ResolvePtr(env, jthis)->crd);
};

/*
 * Class:     hpjava_lang_Location
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=hpjava_lang_Location_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete hpjava_lang_Location_Table.ObjPtr[thecppobj];
    hpjava_lang_Location_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     hpjava_lang_Location
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Location_show
  (JNIEnv * env, jobject jthis, jstring message)
{
  int length=env->GetStringUTFLength(message);

  char * cmessage=new char[length+1];

  const char * jmessagePtr=env->GetStringUTFChars(message, 0);

  for (int i=0; i<length; i++)
    cmessage[i]=*(jmessagePtr+i);
  cmessage[length]='\0';

  env->ReleaseStringUTFChars(message, jmessagePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Location %s, Handle %d, Address %x\n", cmessage, pos, hpjava_lang_Location_ResolvePtr(env, jthis));

  delete [] cmessage;

};
