/*
 * File         : hpjava_lang_Range.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:06:49 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/03/19 17:56:43 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "hpjava_lang_Range.h"
#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;
#include "jctable.h"
extern jctable<Dimension> hpjava_lang_Dimension_Table;

Dimension * hpjava_lang_Dimension_ResolvePtr(JNIEnv *, jobject);
Stride *hpjava_lang_Stride_ResolvePtr(JNIEnv *, jobject);
Location * hpjava_lang_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Range> hpjava_lang_Range_Table;

/*ResolvePtr
 */
Range * hpjava_lang_Range_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(hpjava_lang_Range_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     hpjava_lang_Range
 * Method:    idx
 * Signature: (Lhpjava/lang/Location;)I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_idx
  (JNIEnv * env, jobject jthis, jobject jlocation)
{
  return (hpjava_lang_Range_ResolvePtr(env, jthis)->idx(*hpjava_lang_Location_ResolvePtr(env, jlocation)));
};

/*
 * Class:     hpjava_lang_Range
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Range_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("hpjava/lang/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Range> *temphpjava_lang_Range_Table=new jctable<Range>;
  hpjava_lang_Range_Table=*temphpjava_lang_Range_Table;
  delete temphpjava_lang_Range_Table;
};

/*
 * Class:     hpjava_lang_Range
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Range_clean
  (JNIEnv *, jclass);

/*
 * Class:     hpjava_lang_Range
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Range_finalize
  (JNIEnv * env, jobject jthis)
{
  int thecppobj=env->GetIntField(jthis, hCPPObjID);
  int rc=hpjava_lang_Range_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete hpjava_lang_Range_Table.ObjPtr[thecppobj];
    hpjava_lang_Range_Table.ObjPtr[thecppobj]=NULL;
  }
}

/*
 * Class:     hpjava_lang_Range
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Range_constructor__
  (JNIEnv *env, jobject jthis)
{
  Range *cppRange= new Range();
  long pos=hpjava_lang_Range_Table.NewHandle(cppRange);

  if (hpjava_lang_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_lang_Range_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

}

/*
 * Class:     hpjava_lang_Range
 * Method:    constructor
 * Signature: (I)V
 */
/*
JNIEXPORT void JNICALL Java_hpjava_lang_Range_constructor__I
  (JNIEnv * env, jobject jthis, jint extend)
{
  Range *cppRange= new Range(extend);
  long pos=hpjava_lang_Range_Table.NewHandle(cppRange);

  if (hpjava_lang_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_lang_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);
};

*/

/*
 * Class:     hpjava_lang_Range
 * Method:    constructor
 * Signature: (Lhpjava/lang/Dimension;)V
 */
/*
JNIEXPORT void JNICALL Java_hpjava_lang_Range_constructor__Lhpjava_lang_Dimension_2
  (JNIEnv *env, jobject jthis, jobject dim) 
{
  //  (new Range(*hpjava_lang_Dimension_ResolvePtr(env,dim)))->show();
  Range *cppRange= new Range(*hpjava_lang_Dimension_ResolvePtr(env,dim));
  long pos=hpjava_lang_Range_Table.NewHandle(cppRange);
  
  //hpjava_lang_Dimension_ResolvePtr(env,dim)->show();
  //hpjava_lang_Range_ObjPtr[pos]->show(); //debug;

  if (hpjava_lang_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_lang_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);

}
*/

/*
 * Class:     hpjava_lang_Range
 * Method:    dim
 * Signature: (I)Lhpjava/lang/Dimension;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Range_dim
  (JNIEnv *env, jobject jthis)
{

  //Dimension *newCPPdimPtr=new Dimension();

  Dimension CPPdim = hpjava_lang_Range_ResolvePtr(env, jthis)->dim();
  
  //*newCPPdimPtr=CPPdim;

  //long pos=hpjava_lang_Dimension_NewHandle(newCPPdimPtr);

  //printf("dim pos %ld, address %x\n", pos, hpjava_lang_Dimension_ObjPtr[pos]);
  //printf("is this dim? crd %d\n", hpjava_lang_Dimension_ObjPtr[pos]->crd());

  jclass ccDimension=NULL;
  ccDimension=env->FindClass("hpjava/lang/Dimension");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccDimension, "<init>", "()V");
  jobject jNode=env->NewObject(ccDimension, handleConstructorID);

  Dimension * newCPPdimPtr=hpjava_lang_Dimension_ResolvePtr(env, jNode);

  *newCPPdimPtr=CPPdim;

  return(jNode);
};

/*
 * Class:     hpjava_lang_Range
 * Method:    subrng
 * Signature: (III)Lhpjava/lang/Range;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Range_subrng
  (JNIEnv *env, jobject jthis, jint extent, jint base, jint stride)
{
  Range CPPrange = hpjava_lang_Range_ResolvePtr(env, jthis)->subrng(extent,base,stride);

  jclass ccRange=NULL;
  ccRange=env->FindClass("hpjava/lang/Range");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccRange, "<init>", "()V");
  jobject jNode=env->NewObject(ccRange, handleConstructorID);

  Range * newCPPrangePtr=hpjava_lang_Range_ResolvePtr(env, jNode);

  *newCPPrangePtr=CPPrange;

  return(jNode);
  
}

/*
 * Class:     hpjava_lang_Range
 * Method:    triplet
 * Signature: (III)Lhpjava/lang/Range;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Range_triplet
  (JNIEnv * env, jobject jthis, jint base, jint top, jint stride)
{
  Range CPPrange = hpjava_lang_Range_ResolvePtr(env, jthis)->triplet(base, top, stride);

  jclass ccRange=NULL;
  ccRange=env->FindClass("hpjava/lang/Range");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccRange, "<init>", "()V");
  jobject jNode=env->NewObject(ccRange, handleConstructorID);

  Range * newCPPrangePtr=hpjava_lang_Range_ResolvePtr(env, jNode);

  *newCPPrangePtr=CPPrange;

  return(jNode);
};

/*
 * Class:     hpjava_lang_Range
 * Method:    size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_size
  (JNIEnv * env, jobject jthis)
{
  int temp = hpjava_lang_Range_ResolvePtr(env, jthis)->size();
  //printf("testing on: %d\n", temp);
  //cout<<"testing on: " << temp <<"\n";
  return(temp);
};

/*
 * Class:     hpjava_lang_Range
 * Method:    lev
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_lev
  (JNIEnv * env, jobject jthis)
{
	return(hpjava_lang_Range_ResolvePtr(env, jthis)->lev());
}

/*
 * Class:     hpjava_lang_Range
 * Method:    bas
 * Signature: ()I
 */
JNIEXPORT  jint JNICALL Java_hpjava_lang_Range_bas
  (JNIEnv * env, jobject jthis)
{
  return(hpjava_lang_Range_ResolvePtr(env, jthis)->bas());  
}

/*
 * Class:     hpjava_lang_Range
 * Method:    str
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_str
  (JNIEnv *env, jobject jthis)
{
  return(hpjava_lang_Range_ResolvePtr(env, jthis)->str());
}

/*
 * Class:     hpjava_lang_Range
 * Method:    ker_glb
 * Signature: (I[I)I
 */
/*
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_ker_1glb
  (JNIEnv *env, jobject jthis, jint glb, jintArray sub)
{
  int size=(env->GetArrayLength(sub));
  int * cppsub=new int [size];

  jboolean isCopy;
  jint * jsub;

  jsub=env->GetIntArrayElements(sub, &isCopy);

  for (int i=0; i<size; i++)
    *(cppsub+i)=*(jsub+i);

  env->ReleaseIntArrayElements(sub, jsub, 0);

  return(hpjava_lang_Range_ResolvePtr(env, jthis)->ker_glb(glb,cppsub));
}
*/

/*
 * Class:     hpjava_lang_Range
 * Method:    width
 * Signature: ()I
 */
// JNIEXPORT jint JNICALL Java_hpjava_lang_Range_width
//   (JNIEnv *env, jobject jthis)
// {
//     return(hpjava_lang_Range_ResolvePtr(env, jthis)->width());
// }

/*
 * Class:     hpjava_lang_Range
 * Method:    volume
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_hpjava_lang_Range_volume
  (JNIEnv *env, jobject jthis)
{
  return(hpjava_lang_Range_ResolvePtr(env, jthis)->volume());
}
//inline Stride Range :: stride(int* size) const {
//  const int itemLen = *size ;
//  *size *= volume() ;
//  return Stride(*this, itemLen) ;
//}

/*
 * Class:     hpjava_lang_Range
 * Method:    stride
 * Signature: (I)Lhpjava/lang/Stride;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Range_stride
  (JNIEnv *env, jobject jthis, jint size)
{
  //Stride *newCPPstridePtr=new Stride();

  int ccsize=size;
  Range parent=*hpjava_lang_Range_ResolvePtr(env, jthis);

  Stride CPPstride(parent, ccsize);
  ccsize*=parent.volume();
 
  //printf("range address %x, stride size %d\n",hpjava_lang_Range_ResolvePtr(env, jthis), ccsize); //debug

  //*newCPPstridePtr=CPPstride;

  //long pos=hpjava_lang_Stride_NewHandle(newCPPstridePtr);

  jclass ccStride=NULL;
  ccStride=env->FindClass("hpjava/lang/Stride");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccStride, "<init>", "()V");
  jobject jNode=env->NewObject(ccStride, handleConstructorID);

  jfieldID jsize=env->GetFieldID(ccStride,"size","I");
  env->SetIntField(jNode,jsize,ccsize);

  Stride *newCPPstridePtr=hpjava_lang_Stride_ResolvePtr(env,jNode);
  *newCPPstridePtr=CPPstride;

  return(jNode);
}


/*
 * Class:     hpjava_lang_Range
 * Method:    location
 * Signature: (I)Lhpjava/lang/Location;
 */
JNIEXPORT jobject JNICALL Java_hpjava_lang_Range_location
   (JNIEnv * env, jobject jthis, jint i)
{
 
  Location CPPlocation = (*hpjava_lang_Range_ResolvePtr(env, jthis))(i);
  
  //*newCPPstridePtr=CPPstride;
  
  //long pos=hpjava_lang_Stride_NewHandle(newCPPstridePtr);
  
  jclass ccLocation=NULL;
  ccLocation=env->FindClass("hpjava/lang/Location");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccLocation, "<init>", "()V");
  jobject jNode=env->NewObject(ccLocation, handleConstructorID);
  
  Location *newCPPlocationPtr=hpjava_lang_Location_ResolvePtr(env,jNode);
  
  *newCPPlocationPtr=CPPlocation;
  
  return(jNode);
   
};

/*
 * Class:     hpjava_lang_Range
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_hpjava_lang_Range_show
  (JNIEnv * env, jobject jthis, jstring message)
{
  int length=env->GetStringUTFLength(message);

  char * cmessage=new char[length+1];

  const char * jmessagePtr=env->GetStringUTFChars(message, 0);

  for (int i=0; i<length; i++)
    cmessage[i]=*(jmessagePtr+i);
  cmessage[length]='\0';

  env->ReleaseStringUTFChars(message, jmessagePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Range %s, Handle %d, Address %x\n", cmessage, pos, hpjava_lang_Range_ResolvePtr(env, jthis));

  hpjava_lang_Range_ResolvePtr(env, jthis)->show();
  
  delete [] cmessage;

};
