/*
 * File         : hpjava_spmd_adlib_Shift.C
 * Author       : Guansong Zhang
 * Created      : Tue Nov  4 21:00:00 1997
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1998/09/30 19:47:01 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "hpjava_spmd_adlib_Shift.h"
#include "adlib.h"

DAD * hpjava_lang_DAD_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;
#include "jctable.h"
jctable<Shift> hpjava_spmd_adlib_Shift_Table;

/*ResolvePtr
 */
Shift * hpjava_spmd_adlib_Shift_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(hpjava_spmd_adlib_Shift_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("hpjava/lang/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");  
  jctable<Shift> *temphpjava_spmd_adlib_Shift_Table=new jctable<Shift>;
  hpjava_spmd_adlib_Shift_Table=*temphpjava_spmd_adlib_Shift_Table;
  delete temphpjava_spmd_adlib_Shift_Table;
};

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_clean
  (JNIEnv *, jclass);

Mode int2mode(int i)
{
  switch (i) {
  case 0:
    return CYCL;
    break;

  case 1:
    return EDGE;
    break;
    
  case 2:
    return NONE;
    break;

  default:
    printf("Not defined mode %d\n", i);
    return CYCL;
  }
  return CYCL;
}

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    constructor
 * Signature: (Lhpjava/lang/DAD;Lhpjava/lang/DAD;IIII)V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_constructor
  (JNIEnv *env, jobject jthis, jobject dDAD, jobject sDAD, jint len, jint shift, jint dim, jint mode)
{
  DAD * dDADPtr = hpjava_lang_DAD_ResolvePtr(env, dDAD); 
  DAD * sDADPtr = hpjava_lang_DAD_ResolvePtr(env, sDAD); 

  Shift *cppShift=new Shift(dDADPtr, sDADPtr, len, shift, dim, int2mode(mode));
  long pos=hpjava_spmd_adlib_Shift_Table.NewHandle(cppShift);

  if (hpjava_spmd_adlib_Shift_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //hpjava_spmd_adlib_Shift_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);  

};

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_finalize
  (JNIEnv * env, jobject jthis) 
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=hpjava_spmd_adlib_Shift_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete hpjava_spmd_adlib_Shift_Table.ObjPtr[thecppobj];
    hpjava_spmd_adlib_Shift_Table.ObjPtr[thecppobj]=NULL;
  }

};

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    execute
 * Signature: ([IJ[IJ)V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_execute___3IJ_3IJ
  (JNIEnv * env, jobject jthis, jintArray dst, jlong dOffset, jintArray src, jlong sOffset)
{
  // int dLength=env->GetArrayLength(dst);
  jint * jDA=env->GetIntArrayElements(dst, 0);
    
  // int sLength=env->GetArrayLength(src);
  jint * jSA=env->GetIntArrayElements(src, 0);

  /*
  {//debug
    int i;

    for (i=0; i<dLength; i++) {
      printf("dst %d: %d | ", i, *(jDA+i));
    }
    printf("\n");
    
    for (i=0; i<sLength; i++) {
      printf("src %d: %d | ", i, *(jSA+i));
    }
    printf("\n");
  }
  */

  hpjava_spmd_adlib_Shift_ResolvePtr(env, jthis)->execute(jDA+dOffset, jSA+sOffset); 

  //printf("recorded remap at %x; dOffset %d; sOffset %d\n", hpjava_spmd_adlib_Shift_ResolvePtr(env, jthis), dOffset, sOffset);  //debug

  /*
  {//debug
    int i;

    for (i=0; i<dLength; i++) {
      printf("dst- %d: %d | ", i, *(jDA+i));
    }
    printf("\n");
    
    for (i=0; i<sLength; i++) {
      printf("src- %d: %d | ", i, *(jSA+i));
    }
    printf("\n");
  }
  */

  env->ReleaseIntArrayElements(dst, jDA, 0);
  env->ReleaseIntArrayElements(src, jSA, JNI_ABORT);
};

/*
 * Class:     hpjava_spmd_adlib_Shift
 * Method:    execute
 * Signature: ([FJ[FJ)V
 */
JNIEXPORT void JNICALL Java_hpjava_spmd_adlib_Shift_execute___3FJ_3FJ
  (JNIEnv * env, jobject jthis, jfloatArray dst, jlong dOffset, jfloatArray src, jlong sOffset)
{
  //  jboolean isCopy;

  // int dLength=env->GetArrayLength(dst);
  jfloat * jDA=env->GetFloatArrayElements(dst, 0);
    
  // int sLength=env->GetArrayLength(src);
  jfloat * jSA=env->GetFloatArrayElements(src, 0);

  hpjava_spmd_adlib_Shift_ResolvePtr(env, jthis)->execute(jDA+dOffset, jSA+sOffset); 

  env->ReleaseFloatArrayElements(dst, jDA, 0);
  env->ReleaseFloatArrayElements(src, jSA, JNI_ABORT);
};
