
// Operations on block distributed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRangeRep : public RangeRep {
public :
  BlockRangeRep(const int _extent, RangeRep* _kernel, const int _blockSize,
                const Format _distFormat = DIST_BLOCK) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim(),
                   _distFormat),
          blockSize(_blockSize) {
    kernel->incr() ;

    ker_hi = (extent - 1) / blockSize ;
  }

  ~BlockRangeRep() {
    kernel->decr() ;
  }

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(Block* blk, const int ker_glb) const ;
  void block2(Block* blk, const int ker_glb,
              const int count, const int base) const ;
  void block3(Block* blk, const int ker_glb,
              const int count, const int base, const int stride) const ;

  void crds(int* ker_count, int* ker_bas, int* ker_str) const ;
  void crds2(int* ker_count, int* ker_bas, int* ker_str,
             const int count, const int base) const ;
  void crds3(int* ker_count, int* ker_bas, int* ker_str,
             const int count, const int base, const int stride) const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  int isAligned(RangeRep* x) const ;

  void show() ;
protected :
  int blockSize ;

  int ker_hi ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedBlockRangeRep': Block-distributed range array with   //
//                                ghost areas.                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedBlockRangeRep : public BlockRangeRep {
public :
  ExtendedBlockRangeRep(const int _extent,
                        RangeRep* _kernel, const int _blockSize,
                        const int _wlo, const int _whi) :
          BlockRangeRep(_extent, _kernel, _blockSize), wlo(_wlo), whi(_whi) {}

  RangeRep* shell() const ;

  int disp(const int sub) const ;

  int offset(const Location* i) const ;

  int volume() const ;

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockRangeRep':                                        //
//      Block-distributed range of array with ``packed'' layout.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockRangeRep : public BlockRangeRep {
public :
  PackedBlockRangeRep(const int _extent,
                      RangeRep* _kernel, const int _blockSize,
                      const int _packing,
                      const Format _distFormat = DIST_BLOCK) :
        BlockRangeRep(_extent, _kernel, _blockSize, _distFormat),
        packing(_packing), width((_blockSize + _packing - 1) / _packing) {}

  RangeRep* shell() const ;

  int disp(const int sub) const ;
  int step(const int sub_stp) const ;

  int offset(const Location* i) const ;

  int volume() const ;

protected :
  int packing ;
  int width ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRangeRep': Block-cyclically distributed ranges.   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRangeRep : public BlockRangeRep {
public :
  BlockCyclicRangeRep(const int _extent, RangeRep* _kernel,
                      const int _blockSize) :
          BlockRangeRep(_extent, _kernel, _blockSize, DIST_BLOCK_CYCLIC) {}

  void location(Location* loc, const int glb) const ;

  int offset(const Location* i) const ;

  int volume() const ;

  int isAligned(RangeRep* x) const ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockCyclicRangeRep':                                  //
//      Block-cyclically distributed range of array with ``packed''    //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockCyclicRangeRep : public PackedBlockRangeRep {
public :
  PackedBlockCyclicRangeRep(const int _extent, RangeRep* _kernel,
                            const int _blockSize, const int _packing) :
          PackedBlockRangeRep(_extent, _kernel, _blockSize, _packing,
                              DIST_BLOCK_CYCLIC) {}

  void location(Location* loc, const int glb) const ;

  int offset(const Location* i) const ;

  int volume() const ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Simplified sub-range classes for contiguous subranges.
//
//   Extended block-cyclic subclasses.
//
//   `show' methods for subclasses.

