

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `DAD': Representation of a general array entity.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct DAD {

  // Constructors

  DAD() {}

  DAD(const int _rank, const Group& _group) : rank(_rank), group(_group) {}

  DAD(const int _rank, const Group& _group, Map* _maps) :
      rank(_rank), group(_group), maps(_maps) {}

  // Data members

  int rank ;

  Group group ;

  Map* maps ;

  // inquiries

  int rnk() const {return rank ; }

  const Group &grp() const {return group ; }

  Range rng(const int r) const {return maps [r].rng() ; }

  const Map &map(const int r) const {return maps [r] ; }

  int str(const int r) const {return maps [r].str() ; }

  DimensionSet sig() const {
    DimensionSet signature ;
    for(int r = 0 ; r < rank ; r++)
      signature += maps [r].rng().dim() ;
    return signature ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `alignCheck'                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline void alignCheck(const DAD* a, const DAD* b) {
  if(!a->grp().isEqual(b->grp()))
    AdlibError("alignCheck: Group mismatch") ;
  if(a->rnk() != b->rnk())
    AdlibError("alignCheck: Rank mismatch") ;
  for(int d = 0 ; d < a->rnk() ; d++)
    if(!a->rng(d).isAligned(b->rng(d)))
      AdlibError("alignCheck: Range misalignment") ;
}

inline void alignCheck(const DAD* a, const DAD* b, const int dim) {
  if(!a->grp().isEqual(b->grp()))
    AdlibError("alignCheck: Group mismatch") ;
  if(a->rnk() != b->rnk() + 1)
    AdlibError("alignCheck: Rank mismatch") ;
  int r = 0 ;
  for(int d = 0 ; d < a->rnk() ; d++)
    if(d != dim) {
      if(!a->rng(d).isAligned(b->rng(r)))
	AdlibError("alignCheck: Range misalignment") ;
      r++ ;
    }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `shapeCheck'                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline void shapeCheck(const DAD* a, const DAD* b) {
  if(a->rnk() != b->rnk())
    AdlibError("shapeCheck: Rank mismatch") ;
  for(int d = 0 ; d < a->rnk() ; d++)
    if(a->rng(d).size() != b->rng(d).size())
      AdlibError("shapeCheck: Shape mismatch") ;
}

//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////  Class Map: blockwise mapping operation.                            //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class Map {
//public :
//  Map(const DAD* _arr) : arr(_arr), rank(_arr->rnk()) {
//    exts     = new int [rank] ;
//    step = new int [rank] ;
//  }
//
//  ~Map() {
//    delete [] steps ;
//    delete [] exts ;
//  }
//
//  void execute() {
//    if(arr->grp().member())
//      loop(0, 0, 0) ;
//  }
//
//protected :
//  int rank ;
//  int *exts, *steps ;
//
//  virtual void opMap(const int offset) = 0 ;
//
//private :
//  const DAD* arr ;
//
//  void loop(int offset, const int r) {
//    if(r == rank)
//      opMap(offset) ;
//    else {
//      Map map = arr->map(r) ;
//
//      LocBlocksIndex i(arr->rng(r)) ;
//      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
//        exts  [r] = i.count ;
//        steps [r] = map.step(i.sub_stp) ;
//
//        loop(offset + map.offset(i), r + 1) ;
//      }
//    }
//  }
//} ;

//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////  Class Map2: blockwise mapping operation applied to pair of         //
////              identically distributed arrays.                        //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class Map2 {
//public :
//  Map2(const DAD* _dst, const DAD* _src) :
//           dst(_dst), src(_src), rank(_src->rnk()) {
//    shapeCheck(dst, src) ;
//
//    exts     = new int [rank] ;
//    dstSteps = new int [rank] ;
//    srcSteps = new int [rank] ;
//  }
//
//  ~Map2() {
//    delete [] srcSteps ;
//    delete [] dstSteps ;
//    delete [] exts ;
//  }
//
//  void execute() {
//    if(src->grp().member())
//      loop(0, 0, 0) ;
//  }
//
//protected :
//  int rank ;
//  int *exts, *srcSteps, *dstSteps ;
//
//  virtual void opMap(const int dstOffset, const int srcOffset) = 0 ;
//
//private :
//  const DAD* dst ;
//  const DAD* src ;
//
//  void loop(int dstOffset, int srcOffset, const int r) {
//    if(r == rank)
//      opMap(dstOffset, srcOffset) ;
//    else {
//      Map dstMap = dst->map(r) ;
//      Map srcMap = src->map(r) ;
//
//      LocBlocksIndex i(dst->rng(r)) ;
//      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
//        exts     [r] = i.count ;
//        dstSteps [r] = dstMap.step(i.sub_stp) ;
//        srcSteps [r] = srcMap.step(i.sub_stp) ;
//
//        loop(dstOffset + dstMap.offset(i), srcOffset + srcMap.offset(i),
//             r + 1) ;
//      }
//    }
//  }
//} ;


// ----------------------------------------------------------------------
// Things to do
//
//   Change to (single?) constructor
//
//     DAD(const int _rank)
//
//   which internally allocates the `maps' vector?

