

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `DataSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DataSchedule : public MessSchedule {
protected :
  DataSchedule(const int _elementLen) ;

  void transReq(int locOffset, int remId, int remOffset) {

    // Add a transfer request to the appropriate `locIdxs', `remIdxs' vectors.

    (locIdxs [remId]).add(locOffset) ;
    (remIdxs [remId]).add(remOffset) ;
  }

  void build(TreeSchedule* spanTree) ;

private :
  IndexList* remIdxs ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockDataSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockDataSchedule : public BlockMessSchedule {
protected :
  BlockDataSchedule(const int _elementLen, const int _rank) :
        BlockMessSchedule(_elementLen, _rank) {
    remMessages = new SparseArray<MessQueue>(nProcs) ;
  }

  void transReq(int locOffset, int* locStrs, int locCtg,
                int remOffset, int* remStrs, int remCtg,
                int* exts, int vol, int remId) {

    // Add a transfer request to the appropriate lists.

    const int size = elementLen * vol ;
    if(remId == locId) {
      addBlock(locCopies, locOffset, locStrs, locCtg, size, exts, rank) ;
      addBlock(srvCopies, remOffset, remStrs, remCtg, size, exts, rank) ;
      cpyNum++ ;
    }
    else {
      const int hop = (locId - remId + nProcs) % nProcs ;

      MessQueue* locQueue = locMessages.getElement(hop) ;
      addMessageBlock(locQueue->messages, locQueue->blocks,
		      locOffset, locStrs, locCtg, size, exts, rank) ;
      locQueue->remId = remId ;
      locQueue->hop   = hop   ;

      MessQueue* remQueue = remMessages->getElement(hop) ;
      addMessageBlock(remQueue->messages, remQueue->blocks,
		      remOffset, remStrs, remCtg, size, exts, rank) ;
      remQueue->remId = remId ;
    }
  }

  void build(TreeSchedule* spanTree) {
    sendRequests(spanTree) ;

    delete remMessages ;

    BlockMessSchedule :: build() ;
  }

private :
  SparseArray<MessQueue>* remMessages ;

  void sendRequests(TreeSchedule* spanTree) ;
} ;


// ----------------------------------------------------------------------
// Things to do
//

