
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherSkeleton'                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GatherSkeleton : public DataSchedule {
public :
  GatherSkeleton(const DAD* _loc, const DAD* _rem, const int len,
                 const DAD**_subs, const int** subsData,
                 const DAD* _msk, const int* mskData) ;

private :
  const DAD* loc ;
  const DAD* rem ;
  const DAD** subs ;
  const DAD*  msk ;

  int rank ;

  Group q ;

  void loop(int locOffset,
            const int** subsData, const int* mskData, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapeSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapeSkeleton : public DataSchedule {
public :
  ReshapeSkeleton(const DAD* _loc, const DAD* _rem, const int len) ;
private :
  const DAD* loc ;
  const DAD* rem ;

  int rank ;

  Group q ;

  void loop(int locOffset, int seqIndex, int seqStride, int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePadSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapePadSkeleton : public DataSchedule {
public :
  ReshapePadSkeleton(const DAD* _loc, int _remSize, const DAD* _pad,
                     const int len) ;
private :
  const DAD* loc ;
  const DAD* pad ;

  int rank ;
  int remSize ;

  Group q ;

  void loop(int locOffset, int seqIndex, int seqStride, int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherSkeleton'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherSkeleton : public BlockDataSchedule {
public :
  VecGatherSkeleton(const DAD* _loc, const DAD* _rem, const int len,
                    const DAD** _subs, const int** _subsData) ;

protected :

  int* exts ;
  int* locSteps ;
  int* remSteps ;

private :
  const DAD* loc ;
  const DAD* rem ;
  const DAD** subs ;
  const int** subsData ;

  int rank ;

  static int blockRank(const int rank, const DAD** subs) ;

  void loop(const int locOffset, const int locCtg,
            const int remOffset, const int remCtg,
            const int vol,
            Group remGrp,
            const int r, const int s) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherMskSkeleton'                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherMskSkeleton : public DataSchedule {
public :
  VecGatherMskSkeleton(const DAD* _loc, const DAD* _rem, const int len,
                       const DAD**_subs, const int** subsData,
                       const DAD* _msk, const int* mskData) ;

private :
  const DAD* loc ;
  const DAD* rem ;
  const DAD** subs ;
  const int** subsData ;
  const DAD*  msk ;

  int rank ;

  void loop(const int locOffset, const int remOffset, const int* mskElem,
            Group remGrp, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Gather'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Gather {
public :
  Gather(const DAD* loc, const DAD* rem, const int len,
         const DAD** subs, const int** subsData,
         const DAD* msk, const int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Scatter'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Scatter {
public :
  Scatter(const DAD* loc, const DAD* rem, const int len,
          const DAD** subs, const int** subsData,
          const DAD* msk, const int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.scatter(locDat, remDat) ;
    bcst.execute(remDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ScatterComb'                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class ScatterComb {
public :
  ScatterComb(const DAD* loc, const DAD* rem, const int len,
              const DAD** subs, const int** subsData,
              const DAD* msk, const int* mskData,
              void (*_op)(void*, void*)) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len),
      op(_op) {}

  void execute(void* locDat, void* remDat) {
    skeleton.combine(locDat, remDat, op) ;
    bcst.execute(remDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
  void (*op)(void*, void*) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Reshape'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Reshape {
public :
  Reshape(const DAD* loc, const DAD* rem, const int len) :
              skeleton(loc, rem, len), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
protected :
  ReshapeSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePad'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapePad {
public :
  ReshapePad(const DAD* loc, int remSize, const DAD* pad, const int len) :
      skeleton(loc, remSize, pad, len), bcst(loc, len) {}

  void execute(void* locDat, void* padDat) {
    skeleton.gather(locDat, padDat) ;
    bcst.execute(padDat) ;
  }
protected :
  ReshapePadSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGather'                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGather {
public :
  VecGather(const DAD* loc, const DAD* rem, const int len,
            const DAD** subs, const int** subsData) :
      skeleton(loc, rem, len, subs, subsData), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
private :
  VecGatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecScatter'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecScatter {
public :
  VecScatter(const DAD* loc, const DAD* rem, const int len,
             const DAD** subs, const int** subsData) :
      skeleton(loc, rem, len, subs, subsData), bcst(rem, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.scatter(locDat, remDat) ;
    bcst.execute(remDat) ;
  }
private :
  VecGatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherMsk'                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherMsk {
public :
  VecGatherMsk(const DAD* loc, const DAD* rem, const int len,
                const DAD** subs, const int** subsData,
                const DAD* msk, const int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(remDat) ;
  }
private :
  VecGatherMskSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecScatterMsk'                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecScatterMsk {
public :
  VecScatterMsk(const DAD* loc, const DAD* rem, const int len,
                const DAD** subs, const int** subsData,
                const DAD* msk, const int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.scatter(locDat, remDat) ;
    bcst.execute(remDat) ;
  }
private :
  VecGatherMskSkeleton skeleton ;
  Replicate bcst ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Split `Gather', `Scatter', ... into `Gather', `GatherMsk', `Scatter',
//       `ScatterMsk', ....

