
class Procs ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionRep': representation class for process array     //
//                          dimension.                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionRep {
public :
  int size() const {return extent ; }
  int crd() const {return coord ; }
  DimensionRep(const int _extent, const int _coord, const int _stride,
               const int _axis) :
    extent(_extent), coord(_coord), stride(_stride), mask(1 << _axis) {}

  void show() const {
    AdlibPrintf("DimensionRep :: show: size = %d crd = %d str_prc = %d "
                "dims = %d\n", extent, coord, stride, mask) ;
  }

private :
  int extent ;
  int stride ;
  int coord  ;

  int mask ;

friend class DimensionSet ;        // accesses `mask'
friend class Dimension ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Dimension': the process array dimension.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionSet ;

class Dimension {
public :
  // constructors and destructors

  Dimension() {rep = (DimensionRep*) 0 ; }

  // methods

  int size() const {return rep ? rep->size() : 1 ; }
  int crd()  const {return rep ? rep->crd()  : 0 ; }

  //DimensionSet dims() const ;

  int str_prc() const {return rep ? rep->stride : 0 ; }

  // operators

  operator int() const {return rep != 0 ; }

  // debugging

  void show () const {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Dimension :: show: Null dimension\n") ;
  }

private :
  DimensionRep* rep ;

  void create(const int _extent, const int _coord, const int _stride,
              const int _axis) {
    rep = new DimensionRep(_extent, _coord, _stride, _axis) ;
  }
  void destroy() {delete rep ; }

friend class Procs ;               // accesses `create', `destroy'.
friend class DimensionSet ;        // accesses `rep'
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionSet': A set of process array dimensions.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionSet {
public :
  // constructors and destructors

  DimensionSet() : mask(0) {}

  // methods

  int member(Dimension d) const {return (mask & d.rep->mask) != 0 ; }

  int isEqual(DimensionSet ds) const {
    return ds.mask == mask ;
  }

  int contains(DimensionSet ds) const {
    return (~mask & ds.mask) == 0 ;
  }

  // operators

  operator int() const {return mask ; }

  DimensionSet operator+(DimensionSet ds) const {
    return DimensionSet(mask | ds.mask) ;
  }

  DimensionSet operator-(DimensionSet ds) const {
    return DimensionSet(mask & ~ds.mask) ;
  }

  DimensionSet operator+(Dimension d) const {
    if(d.rep)
      return DimensionSet(mask | d.rep->mask) ;
    else
      return *this ;
  }

  DimensionSet operator-(Dimension d) const {
    if(d.rep)
      return DimensionSet(mask & ~(d.rep->mask)) ;
    else
      return *this ;
  }

  DimensionSet& operator-=(DimensionSet ds) {
    mask &= ~ds.mask ;
    return *this ;
  }

  DimensionSet& operator+=(DimensionSet ds) {
    mask |= ds.mask ;
    return *this ;
  }

  DimensionSet& operator-=(Dimension d) {
    if(d.rep) mask &= ~(d.rep->mask) ;
    return *this ;
  }

  DimensionSet& operator+=(Dimension d) {
    if(d.rep) mask |= d.rep->mask ;
    return *this ;
  }

private :
  DimensionSet(const long _mask) : mask(_mask) {}

  long mask ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Group'.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Procs ;

class Group {
public :
  // constructors

  Group() {}

  Group(const Procs& p) ;

  // methods

  const Procs* prc() const {return procs ; }

  DimensionSet dims() const {return dimSet ; }

  int member() const {return isLocal ; }

  int size() const ;

  int id() const ;

  int id_prc(const int id) const ;
  int id_abs(const int id) const ;

  int lead_prc() const {return leadId ; }
  int lead_abs() const ;

  void restrict(Dimension d, const int coord) ;
  void restrict(Dimension d) ;
  void restrict(DimensionSet ds) ;

  int isEqual(const Group& p) const {
    return p.procs == procs && p.dimSet.isEqual(dimSet) && p.leadId == leadId ;
  }

  int contains(const Group& p) ;

  // debugging

  void show() {
    AdlibPrintf("Group :: show: prc = %p dims = %x lead_prc = %d member = %d\n",
                procs, (int) dimSet, leadId, isLocal) ;
  }

private :
  const Procs* procs ;

  DimensionSet dimSet ;
  int leadId ;
  int isLocal ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Procs {
public :

  // constructors and destructors

  Procs(const int _rank, int* _n, int* subIds = 0) ;
  ~Procs() ;

private :
  Procs() ;  // Only used in initialisation
  Procs(const Procs& p) ;  // forbid copying.

  // methods

public :
  int rnk() const {return rank ; }

  Dimension dim(const int i) const {
    if(i < 0 || i >= rank)
      AdlibError("Procs :: dim: axis out of range\n") ;
    return dimensions [i] ;
  }

  DimensionSet dims() const {return dimSet ; }

  int member() const {return isLocal ; }
  
  int size() const {return mySize ; }

  int id() const {return myId ; }

  int id_abs(const int id) const {return physIds [id] ; }

  int lead_abs() const {return physIds [0] ; }

  const Group& ctx() const {return context ; }

private :
  int rank ;
  Dimension* dimensions ;
  int* physIds ;
  int isLocal ;
  int myId ;
  int mySize ;

  DimensionSet dimSet ;

  Group context ;

friend void AdlibInit_() ;  // accesses `Procs()'
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Group `apg': the process group executing this control thread.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

extern Group apg ;

class ProcTree ;

void spanTree(ProcTree* tree, const Group& p) ;


inline Group :: Group(const Procs& p) :
    procs(&p), dimSet(p.dims()), leadId(0), isLocal(p.member()) {}

inline int Group :: id_abs(const int id) const {
  return procs->id_abs(id_prc(id)) ;
}

inline int Group :: lead_abs() const {
  return procs->id_abs(leadId) ;
}


// ---------------------------------------------------------------------
//
// Things to do...
//
//    Add `Dimension :: dims' returning a singleton `DimensionSet'
//    (or empty set for a null dimension)?
//
//    There's a `Dimension' class in Java AWT.  Is this a problem?

