

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce: reduction of array to scalar.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Reduce : public RedxSchedule {
public :
  Reduce(const DAD* _src) ;
  ~Reduce() ;

  void execute(const int _resLen, void* res) ;

protected :
  int rank ;
  int *exts, *steps ;

  virtual void opZero() = 0 ;
  virtual void opReduce(const int offset) = 0 ;

private :
  const DAD* src ;

  TreeSchedule bcstTree ;

  void loop(const int offset, const int d) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce2: reduction of two arrays to scalar.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Reduce2 : public RedxSchedule {
public :
  Reduce2(const DAD* _src1, const DAD* _src2) ;
  ~Reduce2() ;

  void execute(const int _resLen, void* res) ;

protected :
  int rank ;
  int *exts, *steps1, *steps2 ;

  virtual void opZero() = 0 ;
  virtual void opReduce(const int src1Offset, const int src2Offset) = 0 ;

private :
  const DAD* src1 ;
  const DAD* src2 ;

  TreeSchedule bcstTree ;

  void loop(const int offset1, const int offset2, const int d) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim: Dimensional reduction of array.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReduceDim : public BlockRedxSchedule {
public :
  ReduceDim(const DAD* _res, const int len,
            const DAD* _src, const int _dim) ;
  ~ReduceDim() ;

  void execute(void* resBase) ;

protected :
  int resRank, srcRank ;
  int *exts, *srcSteps, *resSteps ;

  virtual void opZero(const int resOffset) = 0 ;
  virtual void opReduce(const int resOffset, const int srcOffset) = 0 ;

private :
  const DAD* res ;
  const DAD* src ;
  int dim ;

  void loop(const int resOffset, const int resCtg,
            const int srcOffset,
	    const int vol, const int d, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim2: Dimensional reduction of two arrays.             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReduceDim2 : public BlockRedxSchedule {
public :
  ReduceDim2(const DAD* _res, const int len,
             const DAD* _src1, const DAD* _src2, const int _dim) ;
  ~ReduceDim2() ;

  void execute(void* resBase) ;

protected :
  int resRank, srcRank ;
  int *exts, *src1Steps, *src2Steps, *resSteps ;

  virtual void opZero(const int resOffset) = 0 ;
  virtual void opReduce(const int resOffset,
			const int src1Offset, const int src2Offset) = 0 ;
private :
  const DAD* res  ;
  const DAD* src1 ;
  const DAD* src2 ;
  int dim ;

  void loop(const int resOffset, const int resCtg,
            const int src1Offset, const int src2Offset,
            const int vol, const int d, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate: reduction of array to scalar, tagging elements by    //
//                position in source array.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Locate : public Redx2Schedule {
public :
  Locate(const DAD* _src) ;
  ~Locate() ;

  void execute(const int valLen, void* val, int* tags) ;

protected :
  int rank ;
  int *exts, *srcSteps, *blockBases, *blockSteps ;

  virtual void opZero() = 0 ;
  virtual void opReduce(const int srcOffset) = 0 ;

private :
  const DAD* src ;

  TreeSchedule bcstTree ;

  void loop(const int offset, const int d) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate2: reduction of two arrays to scalar, tagging elements //
//                 by position in source arrays.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Locate2 : public Redx2Schedule {
public :
  Locate2(const DAD* _src1, const DAD* _src2) ;
  ~Locate2() ;

  void execute(const int valLen, void* val, int* tags) ;

protected :
  int rank ;
  int *exts, *src1Steps, *src2Steps, *blockBases, *blockSteps ;

  virtual void opZero() = 0 ;
  virtual void opReduce(const int src1Offset, const int src2Offset) = 0 ;

private :
  const DAD* src1 ;
  const DAD* src2 ;

  TreeSchedule bcstTree ;

  void loop(const int offset1, const int offset2, const int d) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim: Dimensional reduction of array, tagging elements  //
//                   by position in reduced dimension of source array. //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocateDim : public BlockRedx2Schedule {
public :
  LocateDim(const DAD* _val, const int len, const DAD* _tag,
            const DAD* _src, const int _dim) ;
  ~LocateDim() ;

  void execute(void* valBase, int* tagBase) ;

protected :
  int resRank, srcRank ;
  int *exts, *srcSteps, *valSteps, *tagSteps ;

  virtual void opZero(const int valOffset, const int tagOffset) = 0 ;
  virtual void opReduce(const int valOffset, const int tagOffset,
                        const int srcOffset,
                        const int blockBase, const int blockStep) = 0 ;
private :
  const DAD* val ;
  const DAD* tag ;
  const DAD* src ;
  int dim ;

  void loop(const int valOffset, const int valCtg,
            const int tagOffset, const int tagCtg,
            const int srcOffset,
            const int vol, const int d, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim2: Dimensional reduction of two arrays, tagging     //
//                    elements by position in reduced dimension of     //
//                    source array.                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocateDim2 : public BlockRedx2Schedule {
public :
  LocateDim2(const DAD* _val, const int len, const DAD* _tag,
             const DAD* _src1, const DAD* _src2, const int _dim) ;
  ~LocateDim2() ;

  void execute(void* valBase, int* tagBase) ;

protected :
  int resRank, srcRank ;
  int *exts, *src1Steps, *src2Steps, *valSteps, *tagSteps ;

  virtual void opZero(const int valOffset, const int tagOffset) = 0 ;
  virtual void opReduce(const int valOffset, const int tagOffset,
                        const int src1Offset, const int src2Offset,
                        const int blockBase, const int blockStep) = 0 ;
private :
  const DAD* val ;
  const DAD* tag ;
  const DAD* src1 ;
  const DAD* src2 ;
  int dim ;

  void loop(const int valOffset, const int valCtg,
            const int tagOffset, const int tagCtg,
            const int src1Offset, const int src2Offset,
            const int vol, const int d, const int r) ;
} ;


// Typed versions...

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce_                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T>
class Reduce_ : public Reduce {
public :
  void execute(S* _res, T* _srcBase) {
    res     = _res ;
    srcBase = _srcBase ;
    Reduce :: execute(sizeof(T), res) ;
  }

protected :
  Reduce_(const DAD* _src) : Reduce(_src) {}

  S* res ;
  T* srcBase ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Reduce2_                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class Reduce2_ : public Reduce2 {
public :
  Reduce2_(const DAD* _src1, const DAD* _src2) : Reduce2(_src1, _src2) {}

  void execute(S* _res, T* _src1Base, U* _src2Base) {
    res      = _res ;
    src1Base = _src1Base ;
    src2Base = _src2Base ;
    Reduce2 :: execute(sizeof(S), res) ;
  }

protected :
  S* res ;
  T* src1Base ;
  U* src2Base ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate_                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T>
class Locate_ : public Locate {
public :
  Locate_(const DAD* _src) : Locate(_src) {}

  void execute(S* _val, int* _tags, T* _srcBase) {
    val     = _val ;
    tags    = _tags ;
    srcBase = _srcBase ;
    Locate :: execute(sizeof(S), val, tags) ;
  }

protected :
  S* val ;
  int* tags ;
  T* srcBase ;
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Locate2_                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class Locate2_ : public Locate2 {
public :
  Locate2_(const DAD* _src1, const DAD* _src2) : Locate2(_src1, _src2) {}

  void execute(S* _val, int* _tags, T* _src1Base, U* _src2Base) {
    val      = _val ;
    tags     = _tags ;
    src1Base = _src1Base ;
    src2Base = _src2Base ;
    Locate2 :: execute(sizeof(S), val, tags) ;
  }

protected :
  S* val ;
  int* tags ;
  T* src1Base ;
  U* src2Base ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim_                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T>
class ReduceDim_ : public ReduceDim {
public :
  ReduceDim_(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim(_res, sizeof(S), _src, _dim) {}

  void execute(S* _resBase, T* _srcBase) {
    resBase = _resBase ;
    srcBase = _srcBase ;
    ReduceDim :: execute(resBase) ;
  }

protected :
  S* resBase ;
  T* srcBase ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ReduceDim2_                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class ReduceDim2_ : public ReduceDim2 {
public :
  ReduceDim2_(const DAD* _res,
              const DAD* _src1, const DAD* _src2, const int _dim) :
        ReduceDim2(_res, sizeof(T), _src1, _src2, _dim) {}

  void execute(S* _resBase, T* _src1Base, U* _src2Base) {
    resBase  = _resBase ;
    src1Base = _src1Base ;
    src2Base = _src2Base ;
    ReduceDim2 :: execute(resBase) ;
  }

protected :
  S* resBase ;
  T* src1Base ;
  U* src2Base ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim_                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T>
class LocateDim_ : public LocateDim {
public :
  LocateDim_(const DAD* _res, const DAD* _loc,
             const DAD* _src, const int _dim) :
      LocateDim(_res, sizeof(S), _loc, _src, _dim) {}

  void execute(S* _valBase, int* _tagBase, T* _srcBase) {
    valBase = _valBase ;
    tagBase = _tagBase ;
    srcBase = _srcBase ;
    LocateDim :: execute(valBase, tagBase) ;
  }

protected :
  S* valBase ;
  int* tagBase ;
  T* srcBase ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class LocateDim2_                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class LocateDim2_ : public LocateDim2 {
public :
  LocateDim2_(const DAD* _res, const DAD* _loc,
              const DAD* _src1, const DAD* _src2,
              const int _dim) :
      LocateDim2(_res, sizeof(S), _loc, _src1, _src2, _dim) {}

  void execute(S* _valBase, int* _tagBase, T* _src1Base, U* _src2Base) {
    valBase  = _valBase ;
    tagBase  = _tagBase ;
    src1Base = _src1Base ;
    src2Base = _src2Base ;
    LocateDim2 :: execute(valBase, tagBase) ;
  }

protected :
  S* valBase ;
  int* tagBase ;
  T* src1Base ;
  U* src2Base ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Replicate: broadcast elements of an array from roots of the  //
//                   groups over which they are replicated.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Replicate : public BlockTreeSchedule {
public :
  Replicate(const DAD* _res, const int len) ;
  ~Replicate() ;

  void execute(void* resBase) ;

protected :
  int rank ;
  int *exts, *steps ;

private :
  const DAD* res ;

  void loop(const int offset, const int ctg, const int vol, const int d) ;
} ;


// ----------------------------------------------------------------------
// Things to do
//
//   `tagBase' argument of LocateDim :; execute' should be `int*'?
//
//   Typed arithmetic reductions should support result of different
//   type to source?

