

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Sum                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Sum : public Reduce_<T, T> {
public :
  Sum(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    *res += *((T*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    blockSum(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumMsk                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumMsk : public Reduce2_<T, T, int> {
public :
  SumMsk(const DAD* src, const DAD* msk) : Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res += *((T*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockSum(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Product                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Product : public Reduce_<T, T> {
public :
  Product(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    *res *= *((T*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset) {
    blockProduct(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductMsk                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductMsk : public Reduce2_<T, T, int> {
public :
  ProductMsk(const DAD* src, const DAD* msk) : Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    *res *= *((T*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockProduct(res, src1Base + srcOffset, steps1,
                      src2Base + mskOffset, steps2, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Maxval                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Maxval : public Reduce_<T, T> {
public :
  Maxval(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) > *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostneg(res) ;
  }

  void opReduce(const int srcOffset) {
    blockMaxval(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalMsk : public Reduce2_<T, T, int> {
public :
  MaxvalMsk(const DAD* src, const DAD* msk) :
                Reduce2_<T, T, int>(src, msk) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) > *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostneg(res) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockMaxval(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Minval                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Minval : public Reduce_<T, T> {
public :
  Minval(const DAD* _src) : Reduce_<T, T>(_src) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) < *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostpos(res) ;
  }

  void opReduce(const int srcOffset) {
    blockMinval(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalMsk : public Reduce2_<T, T, int> {
public :
  MinvalMsk(const DAD* _src, const DAD* _msk) : Reduce2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* buf) {
    if(*((T*) buf) < *res) *res = *((T*) buf) ;
  }

  void opZero() {
    mostpos(res) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockMinval(res, src1Base + srcOffset, steps1, src2Base + mskOffset, steps2,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class All                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class All : public Reduce_<int, int> {
public :
  All(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
    *res = *res && *((int*) buf) ;
  }

  void opZero() {
    *res = 1 ;
  }

  void opReduce(const int srcOffset) {
    blockAnd(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Any                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Any : public Reduce_<int, int> {
public :
  Any(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
    *res = *res || *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    blockOr(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Count                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Count : public Reduce_<int, int> {
public :
  Count(const DAD* _src) : Reduce_<int, int>(_src) {}

private :
  void opCombine(void* buf) {
    *res += *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int srcOffset) {
    blockCount(res, srcBase + srcOffset, steps, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class DotProduct                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
class DotProduct : public Reduce2_<S, T, U> {
public :
  DotProduct(const DAD* _src1, const DAD* _src2) :
                 Reduce2_<S, T, U>(_src1, _src2) {}

private :
  void opCombine(void* buf) {
    *res += *((S*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int src1Offset, const int src2Offset) {
    blockDotProduct(res, src1Base + src1Offset, steps1,
                         src2Base + src2Offset, steps2, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class BoolDotProduct                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BoolDotProduct : public Reduce2_<int, int, int> {
public :
  BoolDotProduct(const DAD* _src1, const DAD* _src2) :
        Reduce2_<int, int, int>(_src1, _src2) {}

private :
  void opCombine(void* buf) {
    *res += *((int*) buf) ;
  }

  void opZero() {
    *res = 0 ;
  }

  void opReduce(const int src1Offset, const int src2Offset) {
    blockBoolDotProduct(res, src1Base + src1Offset, steps1,
                             src2Base + src2Offset, steps2, exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Maxloc                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Maxloc : public Locate_<T, T> {
public :
  Maxloc(const DAD* _src) : Locate_<T, T>(_src) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) < *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostneg(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset) {
    blockMaxloc(val, tags, rank,
                srcBase + srcOffset, srcSteps, blockBases, blockSteps,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocMsk : public Locate2_<T, T, int> {
public :
  MaxlocMsk(const DAD* _src, const DAD* _msk) :
                Locate2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) < *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostneg(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockMaxloc(val, tags, rank,
                src1Base + srcOffset, src1Steps,
                src2Base + mskOffset, src2Steps,
                blockBases, blockSteps,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class Minloc                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class Minloc : public Locate_<T, T> {
public :
  Minloc(const DAD* _src) : Locate_<T, T>(_src) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) > *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostpos(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset) {
    blockMinloc(val, tags, rank,
                srcBase + srcOffset, srcSteps, blockBases, blockSteps,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocMsk : public Locate2_<T, T, int> {
public :
  MinlocMsk(const DAD* _src, const DAD* _msk) :
      Locate2_<T, T, int>(_src, _msk) {}

private :
  void opCombine(void* bufVal, void* bufTag) {
    if(*((T*) bufVal) > *val) return ;
    int i ;
    if(*((T*) bufVal) == *val)
      for(i = rank - 1 ; i >= 0 ; i--)
        if(((int*) bufTag) [i] > tags [i]) return ;
    *val = *((T*) bufVal) ;
    for(i = 0 ; i < rank ; i++)
      tags [i] = ((int*) bufTag) [i] ;
  }

  void opZero() {
    mostpos(val) ;
    for(int i = 0 ; i < rank ; i++)
      mostpos(tags + i) ;
  }

  void opReduce(const int srcOffset, const int mskOffset) {
    blockMinloc(val, tags, rank,
                src1Base + srcOffset, src1Steps,
                src2Base + mskOffset, src2Steps,
                blockBases, blockSteps,
                exts, rank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumDim : public ReduceDim_<T, T> {
public :
  SumDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockZero(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecSum(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class SumDimMsk                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SumDimMsk : public ReduceDim2_<T, T, int> {
public :
  SumDimMsk(const DAD* _res,
            const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockZero(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset,
                const int mskOffset) {
    blockVecSum(resBase + resOffset, resSteps,
		src1Base + srcOffset, src1Steps,
		src2Base + mskOffset, src2Steps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductDim                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductDim : public ReduceDim_<T, T> {
public :
  ProductDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterProduct(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockOne(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecProduct(resBase + resOffset, resSteps,
		    srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class ProductDimMsk                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class ProductDimMsk : public ReduceDim2_<T, T, int> {
public :
  ProductDimMsk(const DAD* _res,
                const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterProduct(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockOne(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    blockVecProduct(resBase + resOffset, resSteps,
		    src1Base + srcOffset, src1Steps,
		    src2Base + mskOffset, src2Steps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalDim : public ReduceDim_<T, T> {
public :
  MaxvalDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterMaxval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockMostneg(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecMaxval(resBase + resOffset, resSteps,
		   srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxvalDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxvalDimMsk : public ReduceDim2_<T, T, int> {
public :
  MaxvalDimMsk(const DAD* _res,
               const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterMaxval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockMostneg(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    blockVecMaxval(resBase + resOffset, resSteps,
		   src1Base + srcOffset, src1Steps,
		   src2Base + mskOffset, src2Steps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalDim : public ReduceDim_<T, T> {
public :
  MinvalDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<T, T>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterMinval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockMostpos(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecMinval(resBase + resOffset, resSteps,
		   srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinvalDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinvalDimMsk : public ReduceDim2_<T, T, int> {
public :
  MinvalDimMsk(const DAD* _res,
               const DAD* _src, const int _dim, const DAD* _msk) :
        ReduceDim2_<T, T, int>(_res, _src, _msk, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterMinval(resBase + resOffset, resStrs, resExts, rank, (T*) buf) ;
  }

  void opZero(const int resOffset) {
    blockMostpos(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset, const int mskOffset) {
    blockVecMinval(resBase + resOffset, resSteps,
		   src1Base + srcOffset, src1Steps,
		   src2Base + mskOffset, src2Steps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AllDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllDim : public ReduceDim_<int, int> {
public :
  AllDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterAnd(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
  }

  void opZero(const int resOffset) {
    blockOne(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecAnd(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class AnyDim                                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AnyDim : public ReduceDim_<int, int> {
public :
  AnyDim(const DAD* _res, const DAD* _src, const int _dim) :
          ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterOr(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
  }

  void opZero(const int resOffset) {
    blockZero(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecOr(resBase + resOffset, resSteps,
		srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class CountDim                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CountDim : public ReduceDim_<int, int> {
public :
  CountDim(const DAD* _res, const DAD* _src, const int _dim) :
           ReduceDim_<int, int>(_res, _src, _dim) {}

private :
  void opCombine(MultiBlock resBlock, void* buf) {
    const int rank      = resBlock [MBLK_RNK] ;
    const int resOffset = resBlock [MBLK_OFF] ;
    int* resStrs        = resBlock + MBLK_STR ;
    int* resExts        = resStrs  + rank     ;
    blockScatterSum(resBase + resOffset, resStrs, resExts, rank, (int*) buf) ;
  }

  void opZero(const int resOffset) {
    blockZero(resBase + resOffset, resSteps, exts, resRank) ;
  }

  void opReduce(const int resOffset, const int srcOffset) {
    blockVecCount(resBase + resOffset, resSteps,
		  srcBase + srcOffset, srcSteps, exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocDim : public LocateDim_<T, T> {
public :
  MaxlocDim(const DAD* _res, const DAD* _loc,
            const DAD* _src, const int _dim, const int _lb = 0) :
      LocateDim_<T, T>(_res, _loc, _src, _dim), lb(_lb) {}

private :
  const int lb ;

  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valStrs        = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagStrs        = tagBlock + MBLK_STR ;
    int* resExts        = valStrs  + rank     ;
    blockScatterMaxtag(valBase + valOffset, valStrs,
                       tagBase + tagOffset, tagStrs, resExts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
  }

  void opZero(const int valOffset, const int tagOffset) {
    blockMostneg(valBase + valOffset, valSteps, exts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, exts, resRank) ;
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset,
                const int blockBase, const int blockStep) {
    blockVecMaxloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   srcBase + srcOffset, srcSteps, blockBase + lb, blockStep,
                   exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MaxlocDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MaxlocDimMsk : public LocateDim2_<T, T, int> {
public :
  MaxlocDimMsk(const DAD* _res, const DAD* _loc,
               const DAD* _src, const int _dim, const DAD* _msk,
               const int _lb = 0) :
      LocateDim2_<T, T, int>(_res, _loc, _src, _msk, _dim), lb(_lb) {}

private :
  const int lb ;

  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valStrs        = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagStrs        = tagBlock + MBLK_STR ;
    int* resExts        = valStrs  + rank     ;
    blockScatterMaxtag(valBase + valOffset, valStrs,
                       tagBase + tagOffset, tagStrs, resExts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
  }

  void opZero(const int valOffset, const int tagOffset) {
    blockMostneg(valBase + valOffset, valSteps, exts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, exts, resRank) ;
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset, const int mskOffset,
                const int blockBase, const int blockStep) {
    blockVecMaxloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   src1Base + srcOffset, src1Steps,
                   src2Base + mskOffset, src2Steps,
                   blockBase + lb, blockStep,
                   exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocDim                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocDim : public LocateDim_<T, T> {
public :
  MinlocDim(const DAD* _res, const DAD* _loc,
            const DAD* _src, const int _dim, const int _lb = 0) :
      LocateDim_<T, T>(_res, _loc, _src, _dim), lb(_lb) {}

private :
  const int lb ;

  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valStrs        = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagStrs        = tagBlock + MBLK_STR ;
    int* resExts        = valStrs  + rank     ;
    blockScatterMintag(valBase + valOffset, valStrs,
                       tagBase + tagOffset, tagStrs, resExts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
  }

  void opZero(const int valOffset, const int tagOffset) {
    blockMostpos(valBase + valOffset, valSteps, exts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, exts, resRank) ;
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset,
                const int blockBase, const int blockStep) {
    blockVecMinloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   srcBase + srcOffset, srcSteps, blockBase + lb, blockStep,
                   exts, resRank) ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Class MinlocDimMsk                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class MinlocDimMsk : public LocateDim2_<T, T, int> {
public :
  MinlocDimMsk(const DAD* _res, const DAD* _loc,
               const DAD* _src, const int _dim, const DAD* _msk,
               const int _lb = 0) :
      LocateDim2_<T, T, int>(_res, _loc, _src, _msk, _dim), lb(_lb) {}

private :
  const int lb ;

  void opCombine(MultiBlock valBlock, MultiBlock tagBlock,
                 void* bufRes, void* bufTag) {
    const int rank      = valBlock [MBLK_RNK] ;
    const int valOffset = valBlock [MBLK_OFF] ;
    int* valStrs        = valBlock + MBLK_STR ;
    const int tagOffset = tagBlock [MBLK_OFF] ;
    int* tagStrs        = tagBlock + MBLK_STR ;
    int* resExts        = valStrs  + rank     ;
    blockScatterMintag(valBase + valOffset, valStrs,
                       tagBase + tagOffset, tagStrs, resExts, rank,
                       *((T**) &bufRes), *((int**) &bufTag)) ;
  }

  void opZero(const int valOffset, const int tagOffset) {
    blockMostpos(valBase + valOffset, valSteps, exts, resRank) ;
    blockMostpos(tagBase + tagOffset, tagSteps, exts, resRank) ;
  }

  void opReduce(const int valOffset, const int tagOffset,
                const int srcOffset, const int mskOffset,
                const int blockBase, const int blockStep) {
    blockVecMinloc(valBase + valOffset, valSteps, tagBase + tagOffset, tagSteps,
                   src1Base + srcOffset, src1Steps,
                   src2Base + mskOffset, src2Steps,
                   blockBase + lb, blockStep,
                   exts, resRank) ;
  }
} ;


// ----------------------------------------------------------------------
// Things to do
//
//   `tagBase' argument of LocateDim :; execute' should be `int*'?
//
//   Typed arithmetic reductions should support result of different
//   type to source?

