

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `TreeSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class TreeSchedule : public ProcTree {
public :
  void synch() ;

  void broadcast(const int resLen, void* res) ;
  void broadcast(const int resLen, void* res, const int srcId) ;

  void broadcast(const int res1Len, void* res1,
                 const int res2Len, void* res2, const int srcId) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RedxSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class RedxSchedule : public TreeSchedule {
public :
  void combine(const int resLen, void* res) ;
protected :
  virtual void opCombine(void* buf) = 0 ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `Redx2Schedule' class.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Redx2Schedule : public TreeSchedule {
public :
  void combine(const int res1Len, void* res1, const int res2Len, void* res2) ;
protected :
  virtual void opCombine(void* buf1, void* buf2) = 0 ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockTreeSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockTreeSchedule : public ProcTree {
public :
  BlockTreeSchedule(const int _elementLen) : elementLen(_elementLen) {}

  void broadcast(void* base) ;

protected :
  int elementLen ;

  void dataReq(const int offset, int* strs, const int contig,
               const int vol, int* exts, const int rank) {
    addMessageBlock(messages, blocks, offset, strs, contig,
                    elementLen * vol, exts, rank) ;
  }

  void clear() {
    messages.clear() ;
    blocks.clear() ;
  }

  MessList messages ;
  MultiBlockList blocks ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRedxSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRedxSchedule : public BlockTreeSchedule {
public :
  BlockRedxSchedule(const int _elementLen) : BlockTreeSchedule(_elementLen) {}

  void combine(void* base) ;
protected :
  virtual void opCombine(MultiBlock resBlock, void* buf) = 0 ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRedx2Schedule' class.                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRedx2Schedule : public BlockTreeSchedule {
public :
  BlockRedx2Schedule(const int element1Len, const int _element2Len) :
      BlockTreeSchedule(element1Len), element2Len(_element2Len) {}

  void dataReq(const int offset1, int* strs1,
               const int contig1, const int vol1,
               const int offset2, int* strs2,
               const int contig2, const int vol2,
               int* exts, const int rank) {
    addMessageBlocks(messages, blocks,
                     offset1, strs1, contig1, elementLen * vol1,
                     offset2, strs2, contig2, element2Len * vol2,
                     exts, rank) ;
  }

  void combine(void* base1, void* base2) ;
protected :
  int element2Len ;

  virtual void opCombine(MultiBlock res1Block, MultiBlock res2Block,
                         void* buf1, void* buf2) = 0 ;
} ;


// ----------------------------------------------------------------------
// Things to do
//
//     Lose `dataReq' and `clear'?

