
#define STATIC_BUFFER_SIZE 262144

typedef double align_t ;

#define ALIGN_SZ sizeof(align_t)
#define ALIGN_UP(n) (ALIGN_SZ * ((n + ALIGN_SZ - 1) / ALIGN_SZ))

extern align_t statBuf [] ;
extern align_t sendBuf [] ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `List': A basic linked list record.                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


template <class T>
//struct List {
//  T item ;
struct List : public T {
  List<T> *next ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `SparseArray': A lightweight sparse array class.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// getElement: creates new entry or returns existing entry for specified
//             index in range 0 .. size - 1.
//
// toList:     returns a linked list containing all created entries, in
//             index order.

// This implementation uses a linked list initially, but if the number
// of entries crosses some threshold it converts to a vector of pointers.

template <class T>
struct SparseRecord : public List<T> {
  int index ;
} ;

template <class T>
class SparseArray {
public :
  SparseArray(const int _size) :
      size(_size), listValid(1), entries(0), list(0), vector(0) {}

  ~SparseArray() {
    if(listValid)
      while(list) {
	SparseRecord<T>* record = list ;
	list = (SparseRecord<T>*) list->next ;
	delete record ;
      }
    else
      for(int i = 0 ; i < size ; i++) {
	SparseRecord<T>* record = vector [i] ;
	if(record)
	  delete record ;
      }
  }

  T* getElement(const int index) {
    if(vector) {
      SparseRecord<T>** element = vector + index ;
      if(!*element) {
	*element = new SparseRecord<T> ;
	listValid = 0 ;
      }
      //return &((*element)->item) ;
      return *element ;
    }
    else {
      SparseRecord<T>* record ;

      if(list) {
	SparseRecord<T>* prev = 0 ;
	SparseRecord<T>* next = list ;
	while(next && next->index < index) {
	  prev = next ;
	  next = (SparseRecord<T>*) next->next ; 
	}

	// next => next->index >= index, prev => prev->index < index

	if(next && next->index == index)
	  record = next ;
	else {
	  record = new SparseRecord<T> ;
	  record->index = index ;
	  record->next  = next ;
	  if(prev)
	    prev->next = record ;
	  else
	    list = record ;

	  entries++ ;
	  if(threshold()) {

	    // Go over to vector representation.

	    SparseRecord<T>** ptr = (vector = new SparseRecord<T>* [size]) ;
	    for(int i = 0 ; i < size ; i++)
	      *(ptr++) = 0 ;

	    SparseRecord<T>* curr = list ;
	    while(curr) {
	      vector [curr->index] = curr ;
	      curr = (SparseRecord<T>*) curr->next ;
	    }
	  }
	}
      }
      else {
	record = new SparseRecord<T> ;
	record->index = index ;
	record->next = 0 ;
	list = record ;
      }
      
      //return &(record->item) ;
      return record ;
    }
  }

  List<T>* toList() {
    if(!listValid) {

      // Must be using vector representation.

      list = 0 ;
      SparseRecord<T>** ptr = vector + size - 1 ;
      for(int i = 0 ; i < size ; i++) {
        SparseRecord<T>* record = *(ptr--) ;
	if(record) {
	  record->next = list ;
	  list = record ;
	}
      }
      listValid = 1 ;
    }
    return list ;
  }
private:
  SparseArray(const SparseArray<T>& a) {}  // Forbid copying

  int threshold() {return entries * entries > size ; }
      // Threshold for conversion from list to vector representation.

  int size, listValid, entries ;
  SparseRecord<T> *list, **vector ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `Vec': A lightweight ``Vector'' class.                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T, int iniCapacity>
class Vec {
public :
  Vec() : n(0), capacity(iniCapacity), vals(iniVals) {}
  ~Vec() {if(vals != iniVals) delete [] vals ; }

  void add(T val) {
    if(n == capacity) extend(2 * capacity) ;
    vals [n] = val ;
    n++ ;
  }

  T* add() {
    if(n == capacity) extend(2 * capacity) ;
    T* result = vals + n ;
    n++ ;
    return result ;
  }

  T* addSome(const int d) {
    const int newN = n + d ;
    if(newN > capacity) {
      const int newCapacity = 2 * capacity ;
      extend(newN > newCapacity ? newN : newCapacity) ;
    }
    T* ptr = vals + n ;
    n = newN ;
    return ptr ;
  }

  void clear() {n = 0 ; }

  T* last() {return vals + n - 1 ; }

  int size() {return n ; }
  T* data() {return vals ; }
private :
  Vec(const Vec<T, iniCapacity>& v) {} ;  // forbid copying

  T iniVals [iniCapacity] ;

  int n, capacity ;
  T* vals ;

  void extend(const int newCapacity) {

    // newCapacity >= n

    T* newVals = new T [newCapacity] ;
    memcpy((char*) newVals, (char*) vals, sizeof(T) * n) ;
    if(vals != iniVals) delete [] vals ;
    vals     = newVals ;
    capacity = newCapacity ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Messages and blocks.                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Mess {
  int size ;    // Size of message in bytes.
  int num  ;    // Number of blocks in message.
} ;

// typedef Vec<Mess, 1> MessList ;         // *** why doesn't this work? ***
class MessList : public Vec<Mess, 1> {} ;  


typedef int* MultiBlock ;

#define MBLK_LEN 0    // Len of this block parametrization (5 + 2 * rank).
#define MBLK_OFF 1    // Offset of start of block from base address
#define MBLK_CTG 2    // Boolean value => block is contiguous in memory.
#define MBLK_SIZ 3    // Number of bytes in block
#define MBLK_RNK 4    // Rank of block.
#define MBLK_STR 5    // First of `rank' strides (followed by `rank' extents).

// NB: Unit of offset and strides is array element (*not* byte).
// NB: If these constants are changed, `addBlock' below must be changed.

typedef Vec<int, 16> MultiBlockList ;


struct MessQueue {
  MessList messages ;
  MultiBlockList blocks ;
  int remId ;
  int hop ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Adding blocks to queues.                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline void addBlock(MultiBlockList &blocks,
	             const int offset, const int* strs, const int contig,
                     const int size,
                     const int* exts, const int rank) {
  const int len = MBLK_STR + 2 * rank ;
  int* block = blocks.addSome(len) ;

  *(block++) = len ;
  *(block++) = offset ;
  *(block++) = contig ;
  *(block++) = size ;
  *(block++) = rank ;
  int i ;
  for(i = 0 ; i < rank ; i++) *(block++) = *(strs++) ;
  for(i = 0 ; i < rank ; i++) *(block++) = *(exts++) ;
}

inline void nextBlock(MultiBlock &block) {
  block += *block ;  // block += block [MBLK_LEN] ;
}

void addMessageBlock(MessList &messages, MultiBlockList &blocks,
                     const int offset, int* strs, const int contig,
                     const int size,
                     int* exts, const int rank) ;

void addMessageBlocks(MessList &messages, MultiBlockList &blocks,
      const int offset1, int* strs1, const int contig1, const int size1,
      const int offset2, int* strs2, const int contig2, const int size2,
      int* exts, const int rank) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    memory block gather/scatter/copy routines.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void gatherOffBuf(void* base,
                  int* offsets, int count, void* buf, int elementLen) ;

void scatterOffBuf(void* base,
                   int* offsets, int count, void* buf, int elementLen) ;

void copyMblkMblk(void* dstBase, MultiBlock dstBlock,
                  void* srcBase, MultiBlock srcBlock, const int elementLen) ;

void gatherMblkBuf(void* locBase, MultiBlock block, void* buffer,
                   const int elementLen) ;

void scatterMblkBuf(void* locBase, MultiBlock block, void* buffer,
                    const int elementLen) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class GlobalEnv.  Base class for various schedules.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GlobalEnv {
protected :
  GlobalEnv() {
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;
  }

  int locId, nProcs ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Change definition of `MultiBlock' MBLK_OFF field to offset in
//   bytes, rather than elements?  Unclear.
//
//   Systemize the naming and commentary for the three lightweight
//   container templates.
//
//   Allocate `SparseRecord' from a pool.
//
//   *** BUG *** `vector' never gets deallocated in `SparseArray'??
