/*
 * File         : AdlibClock.C
 * Author       : Guansong Zhang
 * Created      : Tue Feb  3 16:36:05 1998
 * Revision     : $Revision: 1.4 $
 * Updated      : $Date: 1998/11/01 18:30:58 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <stdlib.h>

#include <stdio.h>

#include <stdarg.h>

#include <string.h>

#ifndef _WINDOWS
#include <unistd.h>
#include <sys/times.h>
#include <sys/param.h>
#endif

#include <limits.h>

extern "C" {
  int gethostname(char *name, int namelen);
}

#include "adlib.h"

#define MAXNUM 5 

void AdlibClock(int jnum) {
  static long CLKTCK = sysconf(_SC_CLK_TCK);

  static struct tms actTimeBuf[MAXNUM+1];
  static clock_t wallTimeBuf[MAXNUM+1];

  static int timeNum=0;

  static char name[25];

  if ((jnum) == 0 ) {
    gethostname (name, 25);
    timeNum=0;
    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);  
  } 
  else {
    timeNum++;

    if (timeNum > MAXNUM ) {
      printf("Maximum time is %d\n", (timeNum-1));
      exit(1);
    }

    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);

    //printf (" %d - %d \n",timeNum, timeNum-(jnum) );
    
    //printf ( " %ld ----- %ld \n", wallTimeBuf[timeNum], wallTimeBuf[timeNum-(jnum)] );
    
    //printf ( " %ld -- \n", CLKTCK);

    //printf ( " %ld -- %ld \n", actTimeBuf[timeNum].tms_utime, actTimeBuf[timeNum].tms_stime );

    if (timeNum < (jnum)) {
      printf ("%s: Rt = %5.3f, Ut = %5.3f, St = %5.3f (CUt: %5.3f, CSt: %5.3f)\n",
              name,
              (float)(wallTimeBuf[timeNum]-wallTimeBuf[0])/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_utime-actTimeBuf[0].tms_utime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_stime-actTimeBuf[0].tms_stime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cutime-actTimeBuf[0].tms_cutime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cstime-actTimeBuf[0].tms_cstime)/CLKTCK);
    }
    else {
      printf ("%s: Rt = %5.3f, Ut = %5.3f, St = %5.3f (CUt: %5.3f, CSt: %5.3f)\n",
              name,
              (float)(wallTimeBuf[timeNum]-wallTimeBuf[timeNum-(jnum)])/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_utime-actTimeBuf[timeNum-(jnum)].tms_utime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_stime-actTimeBuf[timeNum-(jnum)].tms_stime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cutime-actTimeBuf[timeNum-(jnum)].tms_cutime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cstime-actTimeBuf[timeNum-(jnum)].tms_cstime)/CLKTCK);
    }
  }
}

