/*
 * File         : CollapsedRangeRep.C
 * Author       : Bryan Carpenter
 * Revision     : $Revision: 1.12 $
 * Updated      : $Date: 1999/03/31 18:53:19 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

// Operations on collapsed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRangeRep': collapsed template range.                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CollapsedRangeRep :: ker() const {
  return new DimRangeRep() ;
}

RangeRep* CollapsedRangeRep :: shell() const {
  return (RangeRep*) this ;  // cast away `const'
}

void CollapsedRangeRep :: location(Location* loc, const int glb) const {

  loc->sub = glb ;
  loc->crd = 0 ;
}

void CollapsedRangeRep :: block(Block* blk, const int) const {

  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = 0 ;
  blk->glb_bas = 0 ;

  blk->count   = extent ;
}

void CollapsedRangeRep :: block2(Block* blk, const int ker_glb,
                                 const int count, const int base) const {
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = base ;
  blk->glb_bas = 0 ;

  blk->count   = count ;
}

void CollapsedRangeRep :: block3(Block* blk, const int ker_glb,
                                 const int count, const int base,
                                 const int stride) const {
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  blk->sub_bas = base ;
  blk->glb_bas = 0 ;

  blk->count   = count ;
}

void CollapsedRangeRep :: crds(int* ker_count, int* ker_bas,
                               int* ker_str) const {
  *ker_count = 1 ;
  *ker_bas   = 0 ;
  *ker_str   = 1 ;
}

void CollapsedRangeRep :: crds2(int* ker_count, int* ker_bas,
                                int* ker_str,
                                const int count, const int) const {
  *ker_count = 1 ;
  *ker_bas   = 0 ;
  *ker_str   = 1 ;
}

void CollapsedRangeRep :: crds3(int* ker_count, int* ker_bas,
                                int* ker_str,
                                const int, const int, const int) const {
  *ker_count = 1 ;
  *ker_bas   = 0 ;
  *ker_str   = 1 ;
}


int CollapsedRangeRep :: volume() const {return extent ; }

int CollapsedRangeRep :: isAligned(RangeRep* x) const {
  return x == this || 
        (x->format() == DIST_COLLAPSED && x->tem()->size() == extent &&
         x->size() == extent && x->str() == 1) ;
}

void CollapsedRangeRep :: show() {
  AdlibPrintf("CollapsedRangeRep :: show: "
              "size = %d ker = %p lev = %d\n",
              extent, kernel, level) ;
}

MEMORY_POOL_DEFN(CollapsedRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedCollapsedRangeRep',                                //
//         `ExtendedCollapsedSubRangeRep':                             //
//     Collapsed range of array with ghost areas.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int ExtendedCollapsedRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedCollapsedRangeRep :: volume() const {
  return wlo + parSize + whi ;
}

MEMORY_POOL_DEFN(ExtendedCollapsedRangeRep)


// -------------------------------------------------------------
// Things to do
//

