
// Miscellaneous global operations.

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
//#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "Procs.h"
#include "ProcTree.h"
#include "threads.h"
#include "schedule.h"
#include "TreeSchedule.h"
//#include "adtags.h"
#include "Range.h"
#include "apgStack.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Global `Adlib' operations---initialisation and finalisation.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Procs* world ;


// Following two operations are globally collective

void AdlibInit(int& argc, char** (&argv)) {
  MPI_Init(&argc, &argv) ;

  AdlibInit_() ;
}

void AdlibFinalize() {

  AdlibFinalize_() ;

  MPI_Finalize() ;

  exit(0) ;  // shouldn't be necessary, but...
}
 
void AdlibInit_() {
  world = new Procs ;
  apg = *world ;
  apgStack.init() ;

  Server :: threadsInit() ;

  RangeInit() ;
}

void AdlibFinalize_() {

  Server :: threadsFinalize() ;

  delete world ;
}
 
//void AdlibPrintf(char* message, ...) {
//
//  // **** This won't work for floating point numbers!! ****
//
//  va_list ap ;
//
//  va_start(ap, message) ;
//
//  char* substring = message ;
//  int datum = 0 ;  // Initial value ignored, but keeps compiler quiet.
//  for(char* ch = message ; ; ch++)
//    switch(*ch) {
//      case 0 :
//        printf(substring, datum) ;
//        return ;
//      case '%' :
//        {
//          int sublen = ch - substring ;
//          char* subcopy = new char [sublen + 1] ;
//
//          memcpy(subcopy, substring, sublen) ;
//          subcopy [sublen] = 0 ;
//          printf(subcopy, datum) ;
//
//          delete [] subcopy ;
//        }
//        substring = ch ;
//        datum = va_arg(ap, int) ;  // No double-word arguments!
//        break ;
//      default :
//        break ;
//    }
//
//  va_end(ap) ;
//}

static char* findFormat(char* control) {

  // Basically return pointer to first char after '%', but
  // skipping "%%" substrings.

  char* ptr = control - 1 ;
  do {
    ptr++ ;
    if(!(ptr = strchr(ptr, '%'))) return 0 ;
    ptr++ ;
  } while(*ptr == '%') ;
  return ptr ;
}

void AdlibPrintf(char* _control, ...) {

  // Analogue of `printf', but precede output by local process id.

  printf("Process %d: ", AdlibLocId()) ;

  // Copy control string (copy eventually left unchanged, but will use it
  // as a scratch area, and `_control' might be in an unwritable segment).

  // (Later introduction of `control_' is a bug-fix that ought to be
  // be cleaned up.  Need to save this so we can delete it.)

  char* control_ = new char[strlen(_control) + 1] ;

  char* control = control_ ;
  strcpy(control, _control) ;

  // Process the arguments.

  va_list ap ;
  char* ptr ;
   
  va_start(ap, _control) ;

  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int nmods = 0, width, prcsn ;
    do {
      switch(*ptr) {
	case 0 : 

	  // let `printf' report an error, then go home.

	  printf(control) ;
	  delete [] control_ ;
	  return ;
	case '*' :
	  if(nmods == 0) width = va_arg(ap, int) ;
	  if(nmods == 1) prcsn = va_arg(ap, int) ;
	  nmods++ ;
	  ptr++ ;
	  break ;
	case 'c' : {
	  char value = va_arg(ap, char) ;
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, width, value) ;
	  if(nmods == 2) printf(control, width, prcsn, value) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' : {
	  int value = va_arg(ap, int) ;
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, width, value) ;
	  if(nmods == 2) printf(control, width, prcsn, value) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' : {
	  double value = va_arg(ap, double) ;
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, width, value) ;
	  if(nmods == 2) printf(control, width, prcsn, value) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 's' :
	case 'p' :
	case 'n' : {
	  void* value = va_arg(ap, void*) ;
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, width, value) ;
	  if(nmods == 2) printf(control, width, prcsn, value) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	default :
	  ptr++ ;
      }
    } while(scanning) ;
  }

  va_end(ap) ;

  // Print trailing string

  printf(control) ;

  delete [] control_ ;
}

void AdlibError(char* message) {
  printf("%s\n", message) ;
  MPI_Abort(MPI_COMM_WORLD, 1) ;
}


int AdlibNProcs() {
  int size ;
  MPI_Comm_size(MPI_COMM_WORLD, &size) ;

  return size ;
}

int AdlibLocId() {
  int coord ;
  MPI_Comm_rank(MPI_COMM_WORLD, &coord) ;

  return coord ;
}

//void AdlibSend(const int blockSize, char* data, const int dstId) {
//  MPI_Status status ;
//
//  MPI_Isend(data, blockSize, MPI_CHAR, dstId, DAT, MPI_COMM_WORLD, event) ;
//  Server :: waitClientEvent(&status) ;
//}
//
//int AdlibRecv(const int blockSize, char* data, const int srcId) {
//  MPI_Status status ;
//
//  MPI_Irecv(data, blockSize, MPI_CHAR, srcId, DAT, MPI_COMM_WORLD, event) ;
//  Server :: waitClientEvent(&status) ;
//
//  int count ;
//  MPI_Get_count(&status, MPI_CHAR, &count) ;
//
//  return count ;
//}
//
//void AdlibBroadcast(const int blockSize, char* data) {
//  TreeSchedule tree ;
//  spanTree(&tree, apg) ;
//  tree.broadcast(blockSize, data) ;
//}

// -------------------------------------------------------------------
//
// Things to do:
//

