
// The `Server' class, used to simulate background threads.

#include <stdlib.h>
#include <stdio.h>
#include "MPIinclude.h"
#include "threads.h"

int Server :: tableSize ;
MPI_Request* Server :: events ;
Server** Server :: servers ;
int Server :: numThreads ;

MPI_Request* event ;


void Server :: threadsInit() {
  tableSize = 1 ;

  events  = new MPI_Request [tableSize] ;
  servers = new Server* [tableSize] ;

  events [0] = MPI_REQUEST_NULL ;

  :: event = events ;

  numThreads = 1 ;
}

void Server :: threadsFinalize() {
  delete [] servers ;
  delete [] events  ;
}


Server :: Server() {
  if(numThreads == tableSize) {
    // Double size of `events' and `servers' vectors.

    MPI_Request* oldEvents  = events ;
    Server**     oldServers = servers ;
    
    tableSize += tableSize ;

    events  = new MPI_Request [tableSize] ;
    servers = new Server* [tableSize] ;
    
    events [0] = oldEvents  [0] ;

    :: event = events ;

    for(int i = 1 ; i < numThreads ; i++) {
      events  [i] = oldEvents  [i] ;
      servers [i] = oldServers [i] ;

      (servers [i])->event = events + i ;
    }

    delete [] oldEvents  ;
    delete [] oldServers ;
  }

  events  [numThreads] = MPI_REQUEST_NULL ;
  servers [numThreads] = this ;

  event = events + numThreads ;

  numThreads++ ;
}

Server :: ~Server() {
  numThreads-- ;
}


void Server :: waitClientEvent(MPI_Status* status) {
  // Wait until event in client thread succeeds

  int eventId ;

  MPI_Waitany(numThreads, events, &eventId, status) ;
  while(eventId != 0) {
    (servers [eventId])->transition(status) ;

    MPI_Waitany(numThreads, events, &eventId, status) ;
  }
}

void Server :: waitServerTransition() {
  // Wait until a server transition occurs

  int serverId ;
  MPI_Status status ;

  MPI_Waitany(numThreads - 1, events + 1, &serverId, &status) ;
  (servers [serverId + 1])->transition(&status) ;
}

// -------------------------------------------------------------
// Things to do
//
//   Convince myself that typical MPI implementations of non-blocking
//   communication are acceptably efficient.

