
// Schedules for irregular access, based on ARMCI.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
extern "C" {
  #include "armci.h"
}

#include "globals.h"  // debug
#include "threads.h"
#include "adtags.h"
#include "ProcTree.h"
#include "schedule.h"
#include "TreeSchedule.h"
#include "RMSchedule.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RMSchedule' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void RMSchedule :: gather() {

  for(int remId = 0 ; remId < nProcs ; remId++) {

    PointerList* remVec = remPtrs + remId ;
    PointerList* locVec = locPtrs + remId ;

    const int n = remVec->size() ;
    if(n > 0) {

      armci_giov_t desc ;
      desc.src_ptr_array = remVec->data() ;
      desc.dst_ptr_array = locVec->data() ;
      desc.bytes         = elementLen ;
      desc.ptr_array_len = n ;

      ARMCI_GetV(&desc, 1, remId) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRMSchedule' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockRMSchedule :: gather() {

  int* blockT = blocks.data() ;
  void** locPtr = locPtrs.data() ;
  void** remPtr = remPtrs.data() ;

  for(int iblk = 0 ; iblk < numBlk ; iblk++) {

    const int remId        = blockT [MBLKT_REM] ;
    const int numStrLevels = blockT [MBLKT_NST] ;

    int* exts    = blockT + MBLKT_CTG ;  // `contigLen' prepended to vector.
    int* locStrs = exts + numStrLevels + 1 ;
    int* remStrs = locStrs + numStrLevels ;

    //AdlibPrintf("gather: s = %d rem pt=%p st=%d loc pt=%p st=%d sh=%d,%d id=%d\n", numStrLevels, *remPtr, remStrs [0], *locPtr, locStrs [0], exts [0], exts [1], remId) ;  // debug

    ARMCI_GetS(*remPtr, remStrs, *locPtr, locStrs, exts, numStrLevels, remId) ;

    blockT += blockLen ;
    locPtr++ ;
    remPtr++ ;
  }
}


void BlockRMSchedule :: scatter() {

  int* blockT = blocks.data() ;
  void** locPtr = locPtrs.data() ;
  void** remPtr = remPtrs.data() ;

  for(int iblk = 0 ; iblk < numBlk ; iblk++) {

    const int remId        = blockT [MBLKT_REM] ;
    const int numStrLevels = blockT [MBLKT_NST] ;

    int* exts    = blockT + MBLKT_CTG ;  // `contigLen' prepended to vector.
    int* locStrs = exts + numStrLevels + 1 ;
    int* remStrs = locStrs + numStrLevels ;

    ARMCI_PutS(*locPtr, locStrs, *remPtr, remStrs, exts, numStrLevels, remId) ;

    blockT += blockLen ;
    locPtr++ ;
    remPtr++ ;
  }
}


// ----------------------------------------------------------------------
// Things to do
//

