

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Get'                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Get : public BlockRMSchedule {
public :
  Get(const int _locStr [], void* locDat,
      const DAD* _rem, const int remOff, void** remPtrs,
      const int len) ;

protected :

  int* exts ;
  int* locSteps ;
  int* remSteps ;

private :
  const int* locStr ;

  const DAD* rem ;

  int rank ;

  void loop(const int locOffset, const int remOffset,
            const int numStrLevels, const int contigLen,
            const int vol,
            Group remGrp,
            const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapRMSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class RemapRMSkeleton : public CollBlockRMSchedule {
public :
  RemapRMSkeleton(const DAD* _loc, void* locDat,
                  const DAD* _rem, const int remOff, void** remPtrs,
                  const int len) ;

protected :

  int* exts ;
  int* locSteps ;
  int* remSteps ;

private :
  const DAD* loc ;
  const DAD* rem ;

  int rank ;

public: // debug
  TreeSchedule spanTree ;

  void loop(const int locOffset, const int remOffset,
            const int numStrLevels, const int contigLen,
            const int vol,
            Group remGrp,
            const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapRM'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class RemapRM {
public :
  RemapRM(const DAD* loc, void* _locDat,
          const DAD* rem, const int remOff, void** remPtrs,
          const int len) :
      skeleton(loc, _locDat, rem, remOff, remPtrs, len),
      bcst(loc, len), locDat(_locDat) {}

  void execute() {
    skeleton.gather() ;
    bcst.execute(locDat) ;
  }
private :
public :  // debug
  RemapRMSkeleton skeleton ;
  Replicate bcst ;
  void *locDat ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherRMSkeleton'                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GatherRMSkeleton : public CollRMSchedule {
public :
  GatherRMSkeleton(const DAD* _loc, void* locDat,
                   const DAD* _rem, const int remOff, void** remPtrs,
                   const int len,
                   const DAD**_subs, const int** subsDat,
                   const DAD* _msk, const int* mskDat) ;

private :
  const DAD* loc ;
  const DAD* rem ;
  const DAD** subs ;
  const DAD*  msk ;

  int rank ;

  TreeSchedule spanTree ;

  Group q ;

  void loop(int locOffset,
            const int** subsDat, const int* mskDat, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherRM'                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GatherRM {
public :
  GatherRM(const DAD* loc, void* _locDat,
           const DAD* rem, const int remOff, void** remPtrs, const int len,
           const DAD** subs, const int** subsDat,
           const DAD* msk, const int* mskDat) :
      skeleton(loc, _locDat, rem, remOff, remPtrs, len,
               subs, subsDat, msk, mskDat),
      bcst(loc, len), locDat(_locDat) {}

  void execute() {
    skeleton.gather() ;
    bcst.execute(locDat) ;
  }
private :
  GatherRMSkeleton skeleton ;
  Replicate bcst ;
  void *locDat ;
} ;


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `Scatter'                                                     //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class Scatter {
//public :
//  Scatter(const DAD* loc, const DAD* rem, const int len,
//          const DAD** subs, const int** subsData,
//          const DAD* msk, const int* mskData) :
//      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}
//
//  void execute(void* locDat, void* remDat) {
//    skeleton.scatter(locDat, remDat) ;
//    bcst.execute(remDat) ;
//  }
//private :
//  GatherSkeleton skeleton ;
//  Replicate bcst ;
//} ;
//
//
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `ScatterComb'                                                 //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//
//class ScatterComb {
//public :
//  ScatterComb(const DAD* loc, const DAD* rem, const int len,
//              const DAD** subs, const int** subsData,
//              const DAD* msk, const int* mskData,
//              void (*_op)(void*, void*)) :
//      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len),
//      op(_op) {}
//
//  void execute(void* locDat, void* remDat) {
//    skeleton.combine(locDat, remDat, op) ;
//    bcst.execute(remDat) ;
//  }
//private :
//  GatherSkeleton skeleton ;
//  Replicate bcst ;
//  void (*op)(void*, void*) ;
//} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherRMSkeleton'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherRMSkeleton : public CollBlockRMSchedule {
public :
  VecGatherRMSkeleton(const DAD* _loc, void* locDat,
                      const DAD* _rem, const int remOff, void** remPtrs,
                      const int len,
                      const DAD** _subs, const int** _subsData) ;

protected :

  int* exts ;
  int* locSteps ;
  int* remSteps ;

private :
  const DAD* loc ;
  const DAD* rem ;
  const DAD** subs ;
  const int** subsData ;

  int rank ;

  TreeSchedule spanTree ;

  static int blockRank(const int rank, const DAD** subs) ;

  void loop(const int locOffset, const int remOffset,
            const int numStrLevels, const int contigLen,
            const int vol,
            Group remGrp,
            const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherRM'                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherRM {
public :
  VecGatherRM(const DAD* loc, void* _locDat,
              const DAD* rem, const int remOff, void** remPtrs,
              const int len,
              const DAD** subs, const int** subsData) :
      skeleton(loc, _locDat, rem, remOff, remPtrs, len, subs, subsData),
      bcst(loc, len), locDat(_locDat) {}

  void execute() {
    skeleton.gather() ;
    bcst.execute(locDat) ;
  }
private :
  VecGatherRMSkeleton skeleton ;
  Replicate bcst ;
  void *locDat ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecScatterRM'                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecScatterRM {
public :
  VecScatterRM(const DAD* loc, void* _locDat,
               const DAD* rem, const int remOff, void** remPtrs,
               const int len,
               const DAD** subs, const int** subsData) :
      skeleton(loc, _locDat, rem, remOff, remPtrs, len, subs, subsData),
      bcst(loc, len), locDat(_locDat) {}

  void execute() {
    skeleton.scatter() ;
    bcst.execute(locDat) ;
  }
private :
  VecGatherRMSkeleton skeleton ;
  Replicate bcst ;
  void *locDat ;
} ;


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `VecGatherMsk'                                                //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class VecGatherMsk {
//public :
//  VecGatherMsk(const DAD* loc, const DAD* rem, const int len,
//                const DAD** subs, const int** subsData,
//                const DAD* msk, const int* mskData) :
//      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}
//
//  void execute(void* locDat, void* remDat) {
//    skeleton.gather(locDat, remDat) ;
//    bcst.execute(remDat) ;
//  }
//private :
//  VecGatherMskSkeleton skeleton ;
//  Replicate bcst ;
//} ;
//
//
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `VecScatterMsk'                                               //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class VecScatterMsk {
//public :
//  VecScatterMsk(const DAD* loc, const DAD* rem, const int len,
//                const DAD** subs, const int** subsData,
//                const DAD* msk, const int* mskData) :
//      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len) {}
//
//  void execute(void* locDat, void* remDat) {
//    skeleton.scatter(locDat, remDat) ;
//    bcst.execute(remDat) ;
//  }
//private :
//  VecGatherMskSkeleton skeleton ;
//  Replicate bcst ;
//} ;

//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `WriteFaces'                                                  //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class WriteFaces : public BlockMessSchedule {
//public :
//  WriteFaces(const DAD* _src, const int len,
//             const int _dim,
//             const int* _wlo, const int* _whi, const Mode* _mode) ;
//private :
//  const DAD* src ;
//
//  int rank, dim ;
//  const int* wlo ;
//  const int* whi ;
//  const Mode* mode ;
//
//  int* exts ;
//  int* steps ;
//
//  void loop(const int offset, const int ctg, const int vol,
//            const Group& remGrp, const int r) ;
//
//  void recvLoop(const int offset, const int ctg, const int vol,
//                const Group& remGrp, const int r) ;
//
//  void sendLoop(const int offset, const int ctg, const int vol,
//                const Group& remGrp, const int r) ;
//
//} ;
//
//
//class WriteHalo {
//public :
//  WriteHalo(const DAD* src, const int len,
//            const int* wlo, const int* whi, const Mode* mode = 0) :
//          rank(src->rnk()) {
//    faceSchedules = new WriteFaces* [rank] ;
//    for(int r = 0 ; r < rank ; r++)
//      if((mode == 0 || mode [r] != NONE) && (wlo [r] > 0 || whi [r] > 0))
//	faceSchedules [r] = new WriteFaces(src, len, r, wlo, whi, mode) ;
//      else
//	faceSchedules [r] = 0 ;
//  }
//
//  ~WriteHalo() {
//    for(int r = 0 ; r < rank ; r++)
//      if(faceSchedules [r]) delete faceSchedules [r] ;
//    delete [] faceSchedules ;
//  }
//
//  void execute(void* base) {
//    for(int r = 0 ; r < rank ; r++)
//      if(faceSchedules [r]) faceSchedules [r]->gather(base, base) ;
//  }
//
//  void show() {
//    for(int r = 0 ; r < rank ; r++)
//      if(faceSchedules [r]) faceSchedules [r]->show() ;
//  }
//private :
//  const int rank ;
//  WriteFaces** faceSchedules ;
//} ;


// -------------------------------------------------------------
// Things to do
//
//   Split `Gather', `Scatter', ... into `Gather', `GatherMsk', `Scatter',
//       `ScatterMsk', ....

