#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "apgStack.h"

#include "ADAPT.h"
#include "extname.h"
#include "shpf.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Initialisation and finalisation of Adlib.                           //
//                                                                     //
// Go straight into the Fortran run-time library to pick up the values //
// of `argc' and `argv' (if calling program is Fortran) or Fortran 90  //
// initialisation routines (if calling program is C++).                //
//                                                                     //
// There doesn't seem to be a sensible, portable alternative           //
// (this is probably a bug in the MPI definition).                     //
//                                                                     //
// Choice of Fortran or C++ calling program is made at system          //
// configuration time.  In most cases the choice is mandated by which  //
// set of procedures is implemented here.                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#if defined(APO)

// Fortran calling program...

void AD_INIT_ADLIB(int* nProcs) {
  int    argc = 1;
  char  *argv [] = {"a.out", 0};
  AdlibInit(argc, argv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#elif defined(AIX)

// Fortran calling program...

extern "C" {
  extern int p_xargc ;
  extern char** p_xargv ;
}

void AD_INIT_ADLIB(int* nProcs) {
  AdlibInit(p_xargc, p_xargv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#elif defined(SOLARIS)

// Fortran calling program...

extern "C" {
  extern int __xargc ;
  extern char** __xargv ;
}

void AD_INIT_ADLIB(int* nProcs) {
  AdlibInit(__xargc, __xargv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#elif defined(NAG)

// Fortran calling program...

extern "C" {
  extern int f90_argc ;
  extern char** f90_argv ;
}

void AD_INIT_ADLIB(int* nProcs) {
  AdlibInit(f90_argc, f90_argv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

// C++ calling program...

extern "C" {
  void f90_io_init() ;
  void f90_io_finish() ;
}

void ad_f90_init() {
  f90_io_init() ;
}

void ad_f90_finalize() {
  f90_io_finish() ;
}

void ad_init_adlib(int* argc, char*** argv, int* nProcs) {
  AdlibInit(*argc, *argv) ;
  *nProcs = apg.size() ;
}

void ad_finalize_adlib() {
  AdlibFinalize() ;
}

#elif defined(NAG1)

// C++ calling program...

extern "C" {
  void init_io() ;
  void finish_io() ;
}

void ad_f90_init() {
  init_io() ;
}

void ad_f90_finalize() {
  //finish_io() ;  // doesn't terminate??
}

void ad_init_adlib(int* argc, char*** argv, int* nProcs) {
  AdlibInit(*argc, *argv) ;
  *nProcs = apg.size() ;
}

void ad_finalize_adlib() {
  AdlibFinalize() ;
}

#elif defined(EPC)

// Fortran calling program...

extern "C" {
  extern int xargc ;
  extern char** xargv ;
}

void AD_INIT_ADLIB(int* nProcs) {
  AdlibInit(xargc, xargv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#elif defined(DEC)

// Fortran calling program...

extern "C" {
  extern int for__l_argc ;
  extern char** for__a_argv ;
}

void AD_INIT_ADLIB(int* nProcs) {
  AdlibInit(for__l_argc, for__a_argv) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#elif defined(GENERIC)

// C++ calling program...

extern "C" {
  void f90_io_init() ;
  void f90_io_finish() ;
}

void ad_f90_init() {}

void ad_f90_finalize() {}

void ad_init_adlib(int* argc, char*** argv, int* nProcs) {
  AdlibInit(*argc, *argv) ;
  *nProcs = apg.size() ;
}

void ad_finalize_adlib() {
  AdlibFinalize() ;
}

#else

extern "C" {
  extern int  PCRC_IARGC() ;
  extern void PCRC_GETARG(int*, char*) ;
}

#include <strings.h>

static void getArgs(int &argidx, char *argv[]) {
    argidx = 0;
    char *q;
    int argcnt;
    
    argcnt = PCRC_IARGC();
    while (argidx <= argcnt)
    {
        q = (char *)malloc(200);
        fflush(stdout);
        PCRC_GETARG(&argidx, q);
        argv[argidx] = q;
        q = (char *)index(argv[argidx],' ');
        *q = '\0';
        ++argidx;
    }
}

void AD_INIT_ADLIB(int* nProcs) {
  int argidx;
  char * argv[200];
  char **argvp = argv;

  getArgs(argidx, argv);
  AdlibInit(argidx, argvp) ;
  *nProcs = apg.size() ;
}

void AD_FINALIZE_ADLIB() {
  AdlibFinalize() ;
}

#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Error message -- stack length overflow                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void AD_STACK_ERR(int *stack_len) {

  fprintf (stderr, "** RUNTIME ERROR **  Stack overflow.  "
    "Rerun \'shpf\' using \'-stack\' option to specify a "
    "larger stack size (currently %d)\n",
                    *stack_len);
  exit (1);
}

void AD_C_STACK_ERR(int *stack_len) {

  fprintf (stderr, "** RUNTIME ERROR **  Character stack overflow.  "
    "Rerun \'shpf\' using \'-cstack\' option to specify a "
    "larger stack size (currently %d)\n",
                    *stack_len);
  exit (1);
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Process arrays.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Construction and destruction.

struct Ad_ProcessArrayRecord {
  int rank ;
  int shape [7] ;
  Procs* procs ;
} ;

#define MAX_PROCESS_ARRAYS 100

static Ad_ProcessArrayRecord processArrays [MAX_PROCESS_ARRAYS] ;
static int numProcessArrays = 0 ;

void AD_INIT_PROCS(int* handle, int* rank, int* shape) {

  // Scan `processArrays' table for pre-existing process array of the
  // same shape.  Return handle of that if it exists.

  for(*handle = 0 ; *handle < numProcessArrays ; (*handle)++) {
    Ad_ProcessArrayRecord& processArray = processArrays [*handle] ;
    if(*rank == processArray.rank) {
      int i = 0 ;
      while (i < *rank && shape [i] == processArray.shape [i]) i++ ;
      if(i == *rank) return ;
    }
  }

  if(numProcessArrays == MAX_PROCESS_ARRAYS) {
    fprintf(stderr, "** RUNTIME ERROR **  Process array records exhausted\n") ;
    exit(1) ;
  }
  else {
    int i ;

    // Check size of processor array is number of available processors.

    int size = 1 ;
    for (i = 0 ; i < *rank ; i++)
      size *= shape [i] ;

    int asize = apg.size() ;

    if(size != asize && size != 1) {
      fprintf (stderr, "** RUNTIME ERROR **  "
              "Processor array requires %d processes---running on %d\n",
              size, asize) ;
      exit(1) ;
    }

    // Create new process array, and add it to `processArrays' table.

    Ad_ProcessArrayRecord& processArray = processArrays [*handle] ;

    processArray.rank = *rank ;
    for (i = 0 ; i < *rank ; i++)
      processArray.shape [i] = shape [i] ;
    processArray.procs = new Procs(*rank, shape) ;

    numProcessArrays++ ;
  }
}

void AD_DELETE_PROCS(int* handle) {
}

static Procs* ad_ToProcs(int handle) {
  return (processArrays [handle]).procs ;
}

void AD_SPLIT(int* handle, int* subsize, int* subids) {

  // Partition the Active Process Group into a set of subgroups
  // described by 1d process arrays.  Return a handle to the local one.

  // `subsize', `subids' describe the local subgroup.

  if(numProcessArrays == MAX_PROCESS_ARRAYS) {
    fprintf(stderr, "** RUNTIME ERROR **  Process array records exhausted\n") ;
    exit(1) ;
  }
  else {
    // Create new process array, and add it to `processArrays' table.

    // (This process array doesn't correspond to any processor arrangement in
    // the HPF!)

    *handle = numProcessArrays ;
    Ad_ProcessArrayRecord& processArray = processArrays [*handle] ;

    processArray.rank = -1 ;
        // The `Adlib' array is actually rank 1, but must prevent
        // `ad_init_procs_' matching shape and returning a handle to this
        // pseudo processor arrangement.

    processArray.procs = new Procs(1, subsize, subids) ;

    numProcessArrays++ ;
  }
}

int AD_ON(int* handle) {
  apgStack.push(&apg) ;
  Procs* p = ad_ToProcs(*handle) ;
  apg = *p ;
  return p->member() ; 
}

void AD_NO(int* handle) {
  apgStack.pop(&apg) ;
}

void AD_MYID(int* myid) {
  *myid = apg.id() ;
}


// Inquiry functions

void AD_PROC_COORD(int* handle, int* coords) {

  // Returns coordinates of local process.

  Procs* p = ad_ToProcs(*handle) ;

  int rank = p->rnk() ;

  if(rank)
    for(int i = 0 ; i < p->rnk() ; i++)
      if(p->member())
        coords [i] = p->dim(i).crd() ;
      else
        coords [i] = -1 ;
  else
    if(p->member())
      *coords = 0 ;
    else
      *coords = -1 ;
}

void AD_SCALAR_PROC(int* handle, int* coords) {

  // Appearances notwithstanding, this is a test whether we're on
  // the scalar process referenced by `handle'

  Procs* p = ad_ToProcs(*handle) ;

  if(p->member())
    *coords = 0 ;
  else
    *coords = -1 ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Arrays.                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class shpfDAD : public DAD {
public :
  shpfDAD(int* atr) ;
  //shpfDAD(const int dim, int* atr) ;

  Range  rngs [8] ;
  Stride strs [8] ;

  int offset, size ;

  int subsMask [8] ;
} ;


shpfDAD :: shpfDAD(int* atr) {

  // Construct a DAD from shpf attribute vector.

  const int attrs = atr [0] ;

  const int data_rank  = (attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT ;
  const int embed_rank = (attrs & MA_EMBED_RANK_MASK) >> MA_EMBED_RANK_SHIFT ;

  if(attrs & MA_DISTRIB)
    group = *(ad_ToProcs(atr [2])) ;
  else
    group = apg ;

  int* vecs = atr + 3 ;

  int* l   = vecs ;
  int* n   = l   + data_rank ;
  int* s   = n   + data_rank ;
  int* al  = s   + data_rank + data_rank ;
  int* an  = al  + data_rank ;
  int* as  = an  + data_rank ;
  int* tn  = as  + data_rank ;
  int* b   = tn  + data_rank ;
  int* fmt = b   + data_rank ;
  int* d   = fmt + data_rank ;
  int* ep  = d   + data_rank ;
  int* ed  = ep  + embed_rank ;

  rank = 0 ;
  size = 1 ;
  offset = 0 ;

  const Procs* procs = group.prc() ;

  for(int r = 0 ; r < data_rank ; r++) {
//    Range x_tem, x_dat ;
    Range x_dat ;

//    // Set up template range.
    // Set up parent array range.

    switch(fmt [r]) {
      case DFMT_COLLAPSED :
        x_dat = Range(tn [r]) ;        // Collapsed
        break ;
      case DFMT_BLOCK :
        x_dat = ADAPTBlockRange(an [r], al [r], as [r],
                                procs->dim(d [r]), b [r]) ;
                                       // Block
        break ;
      case DFMT_CYCLIC :
        if(b [r] == 1)
          x_dat = ADAPTCyclicRange(an [r], al [r], as [r], procs->dim(d [r])) ;
                                       // Simple cyclic
        else
          x_dat = ADAPTBlockCyclicRange(an [r], al [r], as [r],
                                        procs->dim(d [r]), b [r]) ;
                                       // Block cyclic
        break ;
    }

//    // Set up parent array range.
//
//    if(an [r] == x_tem.size())
//      x_dat = x_tem ;  // trivial alignment to template
//    else
//      x_dat = x_tem.subrng(an [r], al [r], as [r]) ;

    subsMask [rank] = (n [r] == -2) ;

    Stride stride = x_dat.stride(&size) ;
    if(n [r] == -1) {
      Location i(x_dat, l [r]) ;
      group  /= i ;
      offset += stride.offset(i) ;
    }
    else {
      strs [rank] = stride ;
      if((n [r] == x_dat.size() && s [r] > 0) || n [r] == -2)
	rngs [rank] = x_dat ;  // trivial triplet subscript
      else
        rngs [rank] = x_dat.subrng(n [r], l [r], s [r]) ;
      rank++ ;
    }  
  }

  for(int i_emb = 0 ; i_emb < embed_rank ; i_emb++)
    group.restrict(procs->dim(ed [i_emb]), ep [i_emb]) ;

  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// ``Primitive'' communications.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Remap

template<class T>
inline void ad_remap(T* dst_dat, int* dst_atr, T* src_dat, int* src_atr) {
  shpfDAD src(src_atr) ;
  shpfDAD dst(dst_atr) ;

  Remap schedule(&dst, &src, sizeof(T)) ;
  schedule.execute(dst_dat + dst.offset, src_dat + src.offset) ;
}

void AD_REMAP_II(int* dst_dat, int* dst_atr,
                 int* src_dat, int* src_atr) {
  ad_remap(dst_dat, dst_atr, src_dat, src_atr) ;
}

void AD_REMAP_RR(float* dst_dat, int* dst_atr,
                 float* src_dat, int* src_atr) {
  ad_remap(dst_dat, dst_atr, src_dat, src_atr) ;
}

void AD_REMAP_DD(double* dst_dat, int* dst_atr,
                 double* src_dat, int* src_atr) {
  ad_remap(dst_dat, dst_atr, src_dat, src_atr) ;
}

void AD_REMAP_ZZ(COMPLEX* dst_dat, int* dst_atr,
                 COMPLEX* src_dat, int* src_atr) {
  ad_remap(dst_dat, dst_atr, src_dat, src_atr) ;
}

void AD_REMAP_LL(LOGICAL* dst_dat, int* dst_atr,
                 LOGICAL* src_dat, int* src_atr) {
  ad_remap(dst_dat, dst_atr, src_dat, src_atr) ;
}


// Vector-subscript gather

template<class T>
inline void ad_gather(T* dst_dat, int* dst_atr, T* src_dat, int* src_atr,
                      va_list& ap) {
  shpfDAD src(src_atr) ;
  shpfDAD dst(dst_atr) ;

  int i ;
  DAD* subs [8] ;
  const int* subsData [8] ;
  for (i = 0; i < src.rank ; i++)
    if(src.subsMask [i]) {
      int* sub_dat = va_arg(ap, int*) ;
      int* sub_atr = va_arg(ap, int*)  ;

      shpfDAD* sub = new shpfDAD(sub_atr) ;

      subs     [i] = sub ;
      subsData [i] = sub_dat + sub->offset ;
    }
    else
      subs [i] = 0 ;

  VecGather schedule(&dst, &src, sizeof(T), (const DAD**) subs, subsData) ;
  schedule.execute(dst_dat + dst.offset, src_dat + src.offset) ;

  for (i = 0; i < src.rank ; i++)
    if(subs [i]) delete subs [i] ;
}

void AD_GATHER_II(int* dst_dat, int* dst_atr,
                  int* src_dat, int* src_atr,
                  ...) {
  va_list ap ;
  va_start(ap, src_atr) ;
  ad_gather(dst_dat, dst_atr, src_dat, src_atr, ap) ;
  va_end(ap) ;
}

void AD_GATHER_RR(float* dst_dat, int* dst_atr,
                  float* src_dat, int* src_atr,
                  ...) {
  va_list ap ;
  va_start(ap, src_atr) ;
  ad_gather(dst_dat, dst_atr, src_dat, src_atr, ap) ;
  va_end(ap) ;
}

void AD_GATHER_DD(double* dst_dat, int* dst_atr,
                  double* src_dat, int* src_atr,
                  ...) {
  va_list ap ;
  va_start(ap, src_atr) ;
  ad_gather(dst_dat, dst_atr, src_dat, src_atr, ap) ;
  va_end(ap) ;
}

void AD_GATHER_ZZ(COMPLEX* dst_dat, int* dst_atr,
                  COMPLEX* src_dat, int* src_atr,
                  ...) {
  va_list ap ;
  va_start(ap, src_atr) ;
  ad_gather(dst_dat, dst_atr, src_dat, src_atr, ap) ;
  va_end(ap) ;
}

void AD_GATHER_LL(LOGICAL* dst_dat, int* dst_atr,
                  LOGICAL* src_dat, int* src_atr,
                  ...) {
  va_list ap ;
  va_start(ap, src_atr) ;
  ad_gather(dst_dat, dst_atr, src_dat, src_atr, ap) ;
  va_end(ap) ;
}


// Vector-subscript scatter

template<class T>
inline void ad_scatter(T* src_dat, int* src_atr, T* dst_dat, int* dst_atr,
                       va_list& ap) {
  shpfDAD dst(dst_atr) ;
  shpfDAD src(src_atr) ;

  int i ;
  DAD* subs [8] ;
  const int* subsData [8] ;
  for (i = 0; i < dst.rank ; i++)
    if(dst.subsMask [i]) {
      int* sub_dat = va_arg(ap, int*) ;
      int* sub_atr = va_arg(ap, int*)  ;

      shpfDAD* sub = new shpfDAD(sub_atr) ;

      subs     [i] = sub ;
      subsData [i] = sub_dat + sub->offset ;
    }
    else
      subs [i] = 0 ;

  VecScatter schedule(&src, &dst, sizeof(T), (const DAD**) subs, subsData) ;
  schedule.execute(src_dat + src.offset, dst_dat + dst.offset) ;

  for (i = 0; i < dst.rank ; i++)
    if(subs [i]) delete subs [i] ;
}

void AD_SCATTER_II(int* src_dat, int* src_atr,
                   int* dst_dat, int* dst_atr,
                   ...) {
  va_list ap ;
  va_start(ap, dst_atr) ;
  ad_scatter(src_dat, src_atr, dst_dat, dst_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_RR(float* src_dat, int* src_atr,
                   float* dst_dat, int* dst_atr,
                   ...) {
  va_list ap ;
  va_start(ap, dst_atr) ;
  ad_scatter(src_dat, src_atr, dst_dat, dst_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_DD(double* src_dat, int* src_atr,
                   double* dst_dat, int* dst_atr,
                   ...) {
  va_list ap ;
  va_start(ap, dst_atr) ;
  ad_scatter(src_dat, src_atr, dst_dat, dst_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_ZZ(COMPLEX* src_dat, int* src_atr,
                   COMPLEX* dst_dat, int* dst_atr,
                   ...) {
  va_list ap ;
  va_start(ap, dst_atr) ;
  ad_scatter(src_dat, src_atr, dst_dat, dst_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_LL(LOGICAL* src_dat, int* src_atr,
                   LOGICAL* dst_dat, int* dst_atr,
                   ...) {
  va_list ap ;
  va_start(ap, dst_atr) ;
  ad_scatter(src_dat, src_atr, dst_dat, dst_atr, ap) ;
  va_end(ap) ;
}


// Masked Vector-subscript scatter

template<class T>
inline void ad_scatter_mask(T* src_dat, int* src_atr, T* dst_dat, int* dst_atr,
                            int* msk_dat, int* msk_atr, va_list& ap) {
  shpfDAD dst(dst_atr) ;
  shpfDAD src(src_atr) ;
  shpfDAD msk(msk_atr) ;

  int i ;
  DAD* subs [8] ;
  const int* subsData [8] ;
  for (i = 0; i < dst.rank ; i++)
    if(dst.subsMask [i]) {
      int* sub_dat = va_arg(ap, int*) ;
      int* sub_atr = va_arg(ap, int*)  ;

      shpfDAD* sub = new shpfDAD(sub_atr) ;

      subs     [i] = sub ;
      subsData [i] = sub_dat + sub->offset ;
    }
    else
      subs [i] = 0 ;

  VecScatterMsk schedule(&src, &dst, sizeof(T), (const DAD**) subs, subsData,
                         &msk, msk_dat + msk.offset) ;
  schedule.execute(src_dat + src.offset, dst_dat + dst.offset) ;

  for (i = 0; i < dst.rank ; i++)
    if(subs [i]) delete subs [i] ;
}

void AD_SCATTER_MASK_II(int* src_dat, int* src_atr,
                        int* dst_dat, int* dst_atr,
                        int* msk_dat, int* msk_atr,
                        ...) {
  va_list ap ;
  va_start(ap, msk_atr) ;
  ad_scatter_mask(src_dat, src_atr, dst_dat, dst_atr, msk_dat, msk_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_MASK_RR(float* src_dat, int* src_atr,
                        float* dst_dat, int* dst_atr,
                        int* msk_dat, int* msk_atr,
                        ...) {
  va_list ap ;
  va_start(ap, msk_atr) ;
  ad_scatter_mask(src_dat, src_atr, dst_dat, dst_atr, msk_dat, msk_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_MASK_DD(double* src_dat, int* src_atr,
                        double* dst_dat, int* dst_atr,
                        int* msk_dat, int* msk_atr,
                        ...) {
  va_list ap ;
  va_start(ap, msk_atr) ;
  ad_scatter_mask(src_dat, src_atr, dst_dat, dst_atr, msk_dat, msk_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_MASK_ZZ(COMPLEX* src_dat, int* src_atr,
                        COMPLEX* dst_dat, int* dst_atr,
                        int* msk_dat, int* msk_atr,
                        ...) {
  va_list ap ;
  va_start(ap, msk_atr) ;
  ad_scatter_mask(src_dat, src_atr, dst_dat, dst_atr, msk_dat, msk_atr, ap) ;
  va_end(ap) ;
}

void AD_SCATTER_MASK_LL(LOGICAL* src_dat, int* src_atr,
                        LOGICAL* dst_dat, int* dst_atr,
                        int* msk_dat, int* msk_atr,
                        ...) {
  va_list ap ;
  va_start(ap, msk_atr) ;
  ad_scatter_mask(src_dat, src_atr, dst_dat, dst_atr, msk_dat, msk_atr, ap) ;
  va_end(ap) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Other data movements to support F90 array intrinsics.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Spread

template<class T>
inline void ad_spread(T* res_dat, int* res_atr,
                      T* src_dat, int* src_atr,
                      int* dim, int* ncopies) {

  // NB: `ncopies' argument is ignored.  Shape of result is known from its
  //     attribute vector.

  shpfDAD res(res_atr) ;
  shpfDAD src(src_atr) ;

  // Allocate temporary with dimension `dim' deleted

  Range  tmp_rngs [8] ;
  Stride tmp_strs [8] ;
  int tmp_size = 1 ;
  int s = 0 ;
  for(int r = 0 ; r < res.rnk() ; r++)
    if(r != (*dim - 1)) {
      Range x = res.rng(r) ;
      tmp_rngs [s] = x ;
      tmp_strs [s] = x.stride(&tmp_size) ;
      s++ ;
    }
      
  DAD tmp(res.rnk() - 1, res.grp(), tmp_rngs, tmp_strs) ;
  T* tmp_dat = new T [tmp_size] ;

  Remap remap(&tmp, &src, sizeof(T)) ;
  remap.execute(tmp_dat, src_dat) ;

  Spread<T> spread(&res, &tmp, *dim - 1) ;
  spread.execute(res_dat + res.offset, tmp_dat) ;

  delete [] tmp_dat ;
}

void AD_SPREAD_I(int* res_dat, int* res_atr,
                 int* src_dat, int* src_atr,
                 int* dim, int* ncopies) {
  ad_spread(res_dat, res_atr, src_dat, src_atr, dim, ncopies) ;
}

void AD_SPREAD_R(float* res_dat, int* res_atr,
                 float* src_dat, int* src_atr,
                 int* dim, int* ncopies) {
  ad_spread(res_dat, res_atr, src_dat, src_atr, dim, ncopies) ;
}

void AD_SPREAD_D(double* res_dat, int* res_atr,
                 double* src_dat, int* src_atr,
                 int* dim, int* ncopies) {
  ad_spread(res_dat, res_atr, src_dat, src_atr, dim, ncopies) ;
}

void AD_SPREAD_Z(COMPLEX* res_dat, int* res_atr,
                 COMPLEX* src_dat, int* src_atr,
                 int* dim, int* ncopies) {
  ad_spread(res_dat, res_atr, src_dat, src_atr, dim, ncopies) ;
}

void AD_SPREAD_L(LOGICAL* res_dat, int* res_atr,
                 LOGICAL* src_dat, int* src_atr,
                 int* dim, int* ncopies) {
  ad_spread(res_dat, res_atr, src_dat, src_atr, dim, ncopies) ;
}


// Transpose

template<class T>
inline void ad_transpose(T* res_dat, int* res_atr, T* src_dat, int* src_atr) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;

  // Make source DAD with exchanged ranges.

  Range  rngs [2] ;
  Stride strs [2] ;

  rngs [0] = src.rng(1) ;
  rngs [1] = src.rng(0) ;

  strs [0] = src.str(1) ;
  strs [1] = src.str(0) ;

  DAD exc(2, src.grp(), rngs, strs) ;

  // Copy source to destination

  Remap schedule(&res, &exc, sizeof(T)) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_TRANSPOSE_I(int* res_dat, int* res_atr,
                    int* src_dat, int* src_atr) {
  ad_transpose(res_dat, res_atr, src_dat, src_atr) ;
}

void AD_TRANSPOSE_R(float* res_dat, int* res_atr,
                    float* src_dat, int* src_atr) {
  ad_transpose(res_dat, res_atr, src_dat, src_atr) ;
}

void AD_TRANSPOSE_D(double* res_dat, int* res_atr,
                    double* src_dat, int* src_atr) {
  ad_transpose(res_dat, res_atr, src_dat, src_atr) ;
}

void AD_TRANSPOSE_Z(COMPLEX* res_dat, int* res_atr,
                    COMPLEX* src_dat, int* src_atr) {
  ad_transpose(res_dat, res_atr, src_dat, src_atr) ;
}

void AD_TRANSPOSE_L(LOGICAL* res_dat, int* res_atr,
                    LOGICAL* src_dat, int* src_atr) {
  ad_transpose(res_dat, res_atr, src_dat, src_atr) ;
}


// Circular shift (scalar shift)

template<class T>
inline void ad_cshift_s(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                        int* shf, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;

  Shift schedule(&res, &src, sizeof(int), *shf, *dim - 1, CYCL) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_CSHIFT_S_I(int* res_dat, int* res_atr,
                    int* src_dat, int* src_atr,
                    int* shf, int* dim) {
  ad_cshift_s(res_dat, res_atr, src_dat, src_atr, shf, dim) ;
}

void AD_CSHIFT_S_R(float* res_dat, int* res_atr,
                    float* src_dat, int* src_atr,
                    int* shf, int* dim) {
  ad_cshift_s(res_dat, res_atr, src_dat, src_atr, shf, dim) ;
}

void AD_CSHIFT_S_D(double* res_dat, int* res_atr,
                    double* src_dat, int* src_atr,
                    int* shf, int* dim) {
  ad_cshift_s(res_dat, res_atr, src_dat, src_atr, shf, dim) ;
}

void AD_CSHIFT_S_Z(COMPLEX* res_dat, int* res_atr,
                    COMPLEX* src_dat, int* src_atr,
                    int* shf, int* dim) {
  ad_cshift_s(res_dat, res_atr, src_dat, src_atr, shf, dim) ;
}

void AD_CSHIFT_S_L(LOGICAL* res_dat, int* res_atr,
                    LOGICAL* src_dat, int* src_atr,
                    int* shf, int* dim) {
  ad_cshift_s(res_dat, res_atr, src_dat, src_atr, shf, dim) ;
}


// Circular shift (array shift)

template<class T>
inline void ad_cshift_a(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                        int* shf_dat, int* shf_atr, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;
  shpfDAD shf(shf_atr) ;

  Skew schedule(&res, &src, sizeof(T),
                &shf, shf_dat + shf.offset, *dim - 1, CYCL) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_CSHIFT_A_I(int* res_dat, int* res_atr,
                   int* src_dat, int* src_atr,
                   int* shf_dat, int* shf_atr, int* dim) {
  ad_cshift_a(res_dat, res_atr, src_dat, src_atr, shf_dat, shf_atr, dim) ;
}

void AD_CSHIFT_A_R(float* res_dat, int* res_atr,
                   float* src_dat, int* src_atr,
                   int* shf_dat, int* shf_atr, int* dim) {
  ad_cshift_a(res_dat, res_atr, src_dat, src_atr, shf_dat, shf_atr, dim) ;
}

void AD_CSHIFT_A_D(double* res_dat, int* res_atr,
                   double* src_dat, int* src_atr,
                   int* shf_dat, int* shf_atr, int* dim) {
  ad_cshift_a(res_dat, res_atr, src_dat, src_atr, shf_dat, shf_atr, dim) ;
}

void AD_CSHIFT_A_Z(COMPLEX* res_dat, int* res_atr,
                   COMPLEX* src_dat, int* src_atr,
                   int* shf_dat, int* shf_atr, int* dim) {
  ad_cshift_a(res_dat, res_atr, src_dat, src_atr, shf_dat, shf_atr, dim) ;
}

void AD_CSHIFT_A_L(LOGICAL* res_dat, int* res_atr,
                   LOGICAL* src_dat, int* src_atr,
                   int* shf_dat, int* shf_atr, int* dim) {
  ad_cshift_a(res_dat, res_atr, src_dat, src_atr, shf_dat, shf_atr, dim) ;
}


// Edge-off shift (scalar shift, scalar boundary)

template<class T>
inline void ad_eoshift_ss(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                          int* shf, T* bdr, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;

  Init<T> init(&res) ;
  init.execute(res_dat + res.offset, *bdr) ;

  Shift schedule(&res, &src, sizeof(T), *shf, *dim - 1, EDGE) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_EOSHIFT_SS_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* shf, int* bdr, int* dim) {
  ad_eoshift_ss(res_dat, res_atr, src_dat, src_atr, shf, bdr, dim) ;
}

void AD_EOSHIFT_SS_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* shf, float* bdr, int* dim) {
  ad_eoshift_ss(res_dat, res_atr, src_dat, src_atr, shf, bdr, dim) ;
}

void AD_EOSHIFT_SS_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* shf, double* bdr, int* dim) {
  ad_eoshift_ss(res_dat, res_atr, src_dat, src_atr, shf, bdr, dim) ;
}

void AD_EOSHIFT_SS_Z(COMPLEX* res_dat, int* res_atr,
                     COMPLEX* src_dat, int* src_atr,
                     int* shf, COMPLEX* bdr, int* dim) {
  ad_eoshift_ss(res_dat, res_atr, src_dat, src_atr, shf, bdr, dim) ;
}

void AD_EOSHIFT_SS_L(LOGICAL* res_dat, int* res_atr,
                     LOGICAL* src_dat, int* src_atr,
                     int* shf, LOGICAL* bdr, int* dim) {
  ad_eoshift_ss(res_dat, res_atr, src_dat, src_atr, shf, bdr, dim) ;
}


// Edge-off shift (scalar shift, array boundary)

template<class T>
inline void ad_eoshift_sa(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                          int* shf, T* bdr_dat, int* bdr_atr, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;
  shpfDAD bdr(bdr_atr) ;

  Spread<T> spread(&res, &bdr, *dim - 1) ;
  spread.execute(res_dat + res.offset, bdr_dat + bdr.offset) ;

  Shift schedule(&res, &src, sizeof(T), *shf, *dim - 1, EDGE) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_EOSHIFT_SA_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* shf, int* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_sa(res_dat, res_atr, src_dat, src_atr,
                shf, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_SA_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* shf, float* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_sa(res_dat, res_atr, src_dat, src_atr,
                shf, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_SA_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* shf, double* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_sa(res_dat, res_atr, src_dat, src_atr,
                shf, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_SA_Z(COMPLEX* res_dat, int* res_atr,
                     COMPLEX* src_dat, int* src_atr,
                     int* shf, COMPLEX* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_sa(res_dat, res_atr, src_dat, src_atr,
                shf, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_SA_L(LOGICAL* res_dat, int* res_atr,
                     LOGICAL* src_dat, int* src_atr,
                     int* shf, LOGICAL* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_sa(res_dat, res_atr, src_dat, src_atr,
                shf, bdr_dat, bdr_atr, dim) ;
}


// Edge-off shift (array shift, scalar boundary)

template<class T>
inline void ad_eoshift_as(T* res_dat, int* res_atr,
                          T* src_dat, int* src_atr,
                          int* shf_dat, int* shf_atr, T* bdr, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;
  shpfDAD shf(shf_atr) ;

  Init<T> init(&res) ;
  init.execute(res_dat + res.offset, *bdr) ;

  Skew schedule(&res, &src, sizeof(T),
                &shf, shf_dat + shf.offset, *dim - 1, EDGE) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_EOSHIFT_AS_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr, int* bdr, int* dim) {
  ad_eoshift_as(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr, dim) ;
}

void AD_EOSHIFT_AS_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr, float* bdr, int* dim) {
  ad_eoshift_as(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr, dim) ;
}

void AD_EOSHIFT_AS_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr, double* bdr, int* dim) {
  ad_eoshift_as(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr, dim) ;
}

void AD_EOSHIFT_AS_Z(COMPLEX* res_dat, int* res_atr,
                     COMPLEX* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr, COMPLEX* bdr, int* dim) {
  ad_eoshift_as(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr, dim) ;
}

void AD_EOSHIFT_AS_L(LOGICAL* res_dat, int* res_atr,
                     LOGICAL* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr, LOGICAL* bdr, int* dim) {
  ad_eoshift_as(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr, dim) ;
}


// Edge-off shift (array shift, array boundary)

template<class T>
inline void ad_eoshift_aa(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                          int* shf_dat, int* shf_atr,
                          T* bdr_dat, int* bdr_atr, int* dim) {
  shpfDAD src(src_atr) ;
  shpfDAD res(res_atr) ;
  shpfDAD shf(shf_atr) ;
  shpfDAD bdr(bdr_atr) ;

  Spread<T> spread(&res, &bdr, *dim - 1) ;
  spread.execute(res_dat + res.offset, bdr_dat + bdr.offset) ;

  Skew schedule(&res, &src, sizeof(T),
                &shf, shf_dat + shf.offset, *dim - 1, EDGE) ;
  schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
}

void AD_EOSHIFT_AA_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr,
                     int* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_aa(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_AA_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr,
                     float* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_aa(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_AA_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr,
                     double* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_aa(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_AA_Z(COMPLEX* res_dat, int* res_atr,
                     COMPLEX* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr,
                     COMPLEX* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_aa(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr_dat, bdr_atr, dim) ;
}

void AD_EOSHIFT_AA_L(LOGICAL* res_dat, int* res_atr,
                     LOGICAL* src_dat, int* src_atr,
                     int* shf_dat, int* shf_atr,
                     LOGICAL* bdr_dat, int* bdr_atr, int* dim) {
  ad_eoshift_aa(res_dat, res_atr, src_dat, src_atr,
                shf_dat, shf_atr, bdr_dat, bdr_atr, dim) ;
}


// Reshape

template<class T>
inline void ad_reshape(T* dst_dat, int* dst_atr,
                       T* src_dat, int* src_atr,
                       int* shp_dat,  int* shp_atr,
                       T* pad_dat, int* pad_atr, int* pad_present,
                       int* ord_dat, int* ord_atr, int* ord_present) {

  // NB: Shape arguments are ignored (shape of result is known from its
  //     attribute vector).
  //     `order_atr' argument is ignored (`order' is known to be
  //     of global vector of length equal to the rank of `res').

  shpfDAD src(src_atr) ;
  shpfDAD dst(dst_atr) ;

  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  if(*ord_present) {

    // Make target DAD with permuted ranges.

    const int dstRnk = dst.rnk() ;
    Range  rngs [8] ;
    Stride strs [8] ;
    for(int i = 0 ; i < dstRnk ; i++) {
      const int i_tar = ord_dat [i] - 1 ;

      rngs [i_tar] = dst.rng(i) ;
      strs [i_tar] = dst.str(i) ;
        // ... No idea whether I have the permutation the right way round here.
    }

    DAD tar(dstRnk, dst.grp(), rngs, strs) ;

    Reshape schedule(&tar, &src, sizeof(T)) ;
    schedule.execute(dst_dat + dst.offset, src_dat + src.offset) ;

    if(*pad_present) {
      shpfDAD pad(pad_atr) ;

      ReshapePad schedule(&tar, srcSize, &pad, sizeof(T)) ;
      schedule.execute(dst_dat + dst.offset, pad_dat + pad.offset) ;
    }
  }
  else {
    Reshape schedule(&dst, &src, sizeof(T)) ;
    schedule.execute(dst_dat + dst.offset, src_dat + src.offset) ;

    if(*pad_present) {
      shpfDAD pad(pad_atr) ;

      ReshapePad schedule(&dst, srcSize, &pad, sizeof(T)) ;
      schedule.execute(dst_dat + dst.offset, pad_dat + pad.offset) ;
    }
  }
}


void AD_RESHAPE_I(int* res_dat, int* res_atr,
                  int* src_dat, int* src_atr,
                  int* shp_dat, int* shp_atr,
                  int* pad_dat, int* pad_atr, int* pad_present,
                  int* ord_dat, int* ord_atr, int* ord_present) {
  ad_reshape(res_dat, res_atr, src_dat, src_atr,
             shp_dat, shp_atr,
             pad_dat, pad_atr, pad_present, ord_dat, ord_atr, ord_present) ;
}

void AD_RESHAPE_R(float* res_dat, int* res_atr,
                  float* src_dat, int* src_atr,
                  int* shp_dat, int* shp_atr,
                  float* pad_dat, int* pad_atr, int* pad_present,
                  int* ord_dat, int* ord_atr, int* ord_present) {
  ad_reshape(res_dat, res_atr, src_dat, src_atr,
             shp_dat, shp_atr,
             pad_dat, pad_atr, pad_present, ord_dat, ord_atr, ord_present) ;
}

void AD_RESHAPE_D(double* res_dat, int* res_atr,
                  double* src_dat, int* src_atr,
                  int* shp_dat, int* shp_atr,
                  double* pad_dat, int* pad_atr, int* pad_present,
                  int* ord_dat, int* ord_atr, int* ord_present) {
  ad_reshape(res_dat, res_atr, src_dat, src_atr,
             shp_dat, shp_atr,
             pad_dat, pad_atr, pad_present, ord_dat, ord_atr, ord_present) ;
}

void AD_RESHAPE_Z(COMPLEX* res_dat, int* res_atr,
                  COMPLEX* src_dat, int* src_atr,
                  int* shp_dat, int* shp_atr,
                  COMPLEX* pad_dat, int* pad_atr, int* pad_present,
                  int* ord_dat, int* ord_atr, int* ord_present) {
  ad_reshape(res_dat, res_atr, src_dat, src_atr,
             shp_dat, shp_atr,
             pad_dat, pad_atr, pad_present, ord_dat, ord_atr, ord_present) ;
}

void AD_RESHAPE_L(LOGICAL* res_dat, int* res_atr,
                  LOGICAL* src_dat, int* src_atr,
                  int* shp_dat, int* shp_atr,
                  LOGICAL* pad_dat, int* pad_atr, int* pad_present,
                  int* ord_dat, int* ord_atr, int* ord_present) {
  ad_reshape(res_dat, res_atr, src_dat, src_atr,
             shp_dat, shp_atr,
             pad_dat, pad_atr, pad_present, ord_dat, ord_atr, ord_present) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Matrix products.                                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Dot product

// These actually support dot product of any two arrays of the
// same shape and mapping, not just rank-1.

template<class S, class T, class U>
inline void ad_dot_product(S* r, T* a_dat, int* a_atr, U* b_dat, int* b_atr) {
  shpfDAD a(a_atr) ;
  shpfDAD b(b_atr) ;

  if(!(a_atr [0] & MA_DISTRIB) && (b_atr [0] & MA_DISTRIB)) {
    DotProduct<S, U, T> schedule(&b, &a) ;
    schedule.execute(r, b_dat + b.offset, a_dat + a.offset) ;
  }
  else {
    DotProduct<S, T, U> schedule(&a, &b) ;
    schedule.execute(r, a_dat + a.offset, b_dat + b.offset) ;
  }
}

void AD_DOT_PRODUCT_II(int* r,
                       int* a_dat, int* a_atr,
                       int* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_IR(float* r,
                       int* a_dat, int* a_atr,
                       float* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_ID(double* r,
                       int* a_dat, int* a_atr,
                       double* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_IZ(COMPLEX* r,
                       int* a_dat, int* a_atr,
                       COMPLEX* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_RI(float* r,
                       float* a_dat, int* a_atr,
                       int* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_RR(float* r,
                       float* a_dat, int* a_atr,
                       float* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_RD(double* r,
                       float* a_dat, int* a_atr,
                       double* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_RZ(COMPLEX* r,
                       float* a_dat, int* a_atr,
                       COMPLEX* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_DI(double* r,
                       double* a_dat, int* a_atr,
                       int* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_DR(double* r,
                       double* a_dat, int* a_atr,
                       float* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_DD(double* r,
                       double* a_dat, int* a_atr,
                       double* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_DZ(COMPLEX* r,
                       double* a_dat, int* a_atr,
                       COMPLEX* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_ZI(COMPLEX* r,
                       COMPLEX* a_dat, int* a_atr,
                       int* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_ZR(COMPLEX* r,
                       COMPLEX* a_dat, int* a_atr,
                       float* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_ZD(COMPLEX* r,
                       COMPLEX* a_dat, int* a_atr,
                       double* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_ZZ(COMPLEX* r,
                       COMPLEX* a_dat, int* a_atr,
                       COMPLEX* b_dat, int* b_atr) {
  ad_dot_product(r, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_DOT_PRODUCT_LL(int* r,
                       int* a_dat, int* a_atr,
                       int* b_dat, int* b_atr) {
  shpfDAD a(a_atr) ;
  shpfDAD b(b_atr) ;

  if(!(a_atr [0] & MA_DISTRIB) && (b_atr [0] & MA_DISTRIB)) {
    BoolDotProduct schedule(&b, &a) ;
    schedule.execute(r, b_dat + b.offset, a_dat + a.offset) ;
  }
  else {
    BoolDotProduct schedule(&a, &b) ;
    schedule.execute(r, a_dat + a.offset, b_dat + b.offset) ;
  }
}


// Matrix product

template<class S, class T, class U>
inline void ad_matmul(S* res_dat, int* res_atr,
                      T* src1_dat, int* src1_atr,
                      U* src2_dat, int* src2_atr) {
  shpfDAD res(res_atr) ;
  shpfDAD src1(src1_atr) ;
  shpfDAD src2(src2_atr) ;

  matmul(&res, res_dat + res.offset,
         &src1, src1_dat + src1.offset, &src2, src2_dat + src2.offset) ;
}

void AD_MATMUL_II(int* r_dat, int* r_atr,
                  int* a_dat, int* a_atr, int* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_IR(float* r_dat, int* r_atr,
                  int* a_dat, int* a_atr, float* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_ID(double* r_dat, int* r_atr,
                  int* a_dat, int* a_atr, double* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_IZ(COMPLEX* r_dat, int* r_atr,
                  int* a_dat, int* a_atr, COMPLEX* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_RI(float* r_dat, int* r_atr,
                  float* a_dat, int* a_atr, int* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_RR(float* r_dat, int* r_atr,
                  float* a_dat, int* a_atr, float* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_RD(double* r_dat, int* r_atr,
                  float* a_dat, int* a_atr, double* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_RZ(COMPLEX* r_dat, int* r_atr,
                  float* a_dat, int* a_atr, COMPLEX* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_DI(double* r_dat, int* r_atr,
                  double* a_dat, int* a_atr, int* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_DR(double* r_dat, int* r_atr,
                  double* a_dat, int* a_atr, float* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_DD(double* r_dat, int* r_atr,
                  double* a_dat, int* a_atr, double* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_DZ(COMPLEX* r_dat, int* r_atr,
                  double* a_dat, int* a_atr, COMPLEX* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_ZI(COMPLEX* r_dat, int* r_atr,
                  COMPLEX* a_dat, int* a_atr, int* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_ZR(COMPLEX* r_dat, int* r_atr,
                  COMPLEX* a_dat, int* a_atr, float* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_ZD(COMPLEX* r_dat, int* r_atr,
                  COMPLEX* a_dat, int* a_atr, double* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_ZZ(COMPLEX* r_dat, int* r_atr,
                  COMPLEX* a_dat, int* a_atr, COMPLEX* b_dat, int* b_atr) {
  ad_matmul(r_dat, r_atr, a_dat, a_atr, b_dat, b_atr) ;
}

void AD_MATMUL_LL(LOGICAL* r_dat, int* r_atr,
                  LOGICAL* a_dat, int* a_atr, LOGICAL* b_dat, int* b_atr) {
  shpfDAD r(r_atr) ;
  shpfDAD a(a_atr) ;
  shpfDAD b(b_atr) ;

  boolMatmul(&r, r_dat + r.offset,
             &a, a_dat + a.offset, &b, b_dat + b.offset) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Reductions.                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Boolean

void AD_ALL(int* r, int* msk_dat, int* msk_atr) {
  shpfDAD msk(msk_atr) ;
  All schedule(&msk) ;
  schedule.execute(r, msk_dat + msk.offset) ;
}

void AD_ANY(int* r, int* msk_dat, int* msk_atr) {
  shpfDAD msk(msk_atr) ;
  Any schedule(&msk) ;
  schedule.execute(r, msk_dat + msk.offset) ;
}

void AD_COUNT(int* r, int* msk_dat, int* msk_atr) {
  shpfDAD msk(msk_atr) ;
  Count schedule(&msk) ;
  schedule.execute(r, msk_dat + msk.offset) ;
}

void AD_ALL_DIM(int* r_dat, int* r_atr, int* msk_dat, int* msk_atr,
                int* dim) {
  shpfDAD r(r_atr) ;
  shpfDAD msk(msk_atr) ;
  AllDim schedule(&r, &msk, *dim - 1) ;
  schedule.execute(r_dat + r.offset, msk_dat + msk.offset) ;
}

void AD_ANY_DIM(int* r_dat, int* r_atr, int* msk_dat, int* msk_atr,
                int* dim) {
  shpfDAD r(r_atr) ;
  shpfDAD msk(msk_atr) ;
  AnyDim schedule(&r, &msk, *dim - 1) ;
  schedule.execute(r_dat + r.offset, msk_dat + msk.offset) ;
}

void AD_COUNT_DIM(int* r_dat, int* r_atr, int* msk_dat, int* msk_atr,
                  int* dim) {
  shpfDAD r(r_atr) ;
  shpfDAD msk(msk_atr) ;
  CountDim schedule(&r, &msk, *dim - 1) ;
  schedule.execute(r_dat + r.offset, msk_dat + msk.offset) ;
}


// Scalar-valued reductions with masking

// Maximum value

template<class T>
inline void ad_maxval(T* res, T* src_dat, int* src_atr,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MaxvalMsk<T> schedule(&src, &msk) ;
    schedule.execute(res, src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    Maxval<T> schedule(&src) ;
    schedule.execute(res, src_dat + src.offset) ;
  }
}

void AD_MAXVAL_I(int* res, int* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_MAXVAL_R(float* res, float* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_MAXVAL_D(double* res, double* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}


// Minimum value

template<class T>
inline void ad_minval(T* res, T* src_dat, int* src_atr,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MinvalMsk<T> schedule(&src, &msk) ;
    schedule.execute(res, src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    Minval<T> schedule(&src) ;
    schedule.execute(res, src_dat + src.offset) ;
  }
}

void AD_MINVAL_I(int* res, int* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_MINVAL_R(float* res, float* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_MINVAL_D(double* res, double* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}


// Sum

template<class T>
inline void ad_sum(T* res, T* src_dat, int* src_atr,
                   int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    SumMsk<T> schedule(&src, &msk) ;
    schedule.execute(res, src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    Sum<T> schedule(&src) ;
    schedule.execute(res, src_dat + src.offset) ;
  }
}

void AD_SUM_I(int* res, int* src_dat, int* src_atr,
              int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_R(float* res, float* src_dat, int* src_atr,
              int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_D(double* res, double* src_dat, int* src_atr,
              int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_Z(COMPLEX* res, COMPLEX* src_dat, int* src_atr,
              int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}


// Product

template<class T>
inline void ad_product(T* res, T* src_dat, int* src_atr,
                       int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    ProductMsk<T> schedule(&src, &msk) ;
    schedule.execute(res, src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    Product<T> schedule(&src) ;
    schedule.execute(res, src_dat + src.offset) ;
  }
}

void AD_PRODUCT_I(int* res, int* src_dat, int* src_atr,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_R(float* res, float* src_dat, int* src_atr,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_D(double* res, double* src_dat, int* src_atr,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_Z(COMPLEX* res, COMPLEX* src_dat, int* src_atr,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product(res, src_dat, src_atr, msk_dat, msk_atr, msk_present) ;
}


// Array-valued reductions with masking

// Maximum value

template<class T>
inline void ad_maxval_dim(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                          int* dim,
                          int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD res(res_atr) ;
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MaxvalDimMsk<T> schedule(&res, &src, *dim - 1, &msk) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    MaxvalDim<T> schedule(&res, &src, *dim - 1) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
  }
}

void AD_MAXVAL_DIM_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}

void AD_MAXVAL_DIM_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}

void AD_MAXVAL_DIM_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}


// Minimum value

template<class T>
inline void ad_minval_dim(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                          int* dim,
                          int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD res(res_atr) ;
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MinvalDimMsk<T> schedule(&res, &src, *dim - 1, &msk) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    MinvalDim<T> schedule(&res, &src, *dim - 1) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
  }
}

void AD_MINVAL_DIM_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}

void AD_MINVAL_DIM_R(float* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}

void AD_MINVAL_DIM_D(double* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minval_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}


// Sum

template<class T>
inline void ad_sum_dim(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                       int* dim,
                       int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD res(res_atr) ;
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    SumDimMsk<T> schedule(&res, &src, *dim - 1, &msk) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    SumDim<T> schedule(&res, &src, *dim - 1) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
  }
}

void AD_SUM_DIM_I(int* res_dat, int* res_atr,
                  int* src_dat, int* src_atr,
                  int* dim,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum_dim(res_dat, res_atr, src_dat, src_atr,
             dim, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_DIM_R(float* res_dat, int* res_atr,
                  float* src_dat, int* src_atr,
                  int* dim,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum_dim(res_dat, res_atr, src_dat, src_atr,
             dim, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_DIM_D(double* res_dat, int* res_atr,
                  double* src_dat, int* src_atr,
                  int* dim,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum_dim(res_dat, res_atr, src_dat, src_atr,
             dim, msk_dat, msk_atr, msk_present) ;
}

void AD_SUM_DIM_Z(COMPLEX* res_dat, int* res_atr,
                  COMPLEX* src_dat, int* src_atr,
                  int* dim,
                  int* msk_dat, int* msk_atr, int* msk_present) {
  ad_sum_dim(res_dat, res_atr, src_dat, src_atr,
             dim, msk_dat, msk_atr, msk_present) ;
}


// Product

template<class T>
inline void ad_product_dim(T* res_dat, int* res_atr, T* src_dat, int* src_atr,
                           int* dim,
                           int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD res(res_atr) ;
  shpfDAD src(src_atr) ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    ProductDimMsk<T> schedule(&res, &src, *dim - 1, &msk) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    ProductDim<T> schedule(&res, &src, *dim - 1) ;
    schedule.execute(res_dat + res.offset, src_dat + src.offset) ;
  }
}

void AD_PRODUCT_DIM_I(int* res_dat, int* res_atr,
                      int* src_dat, int* src_atr,
                      int* dim,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product_dim(res_dat, res_atr, src_dat, src_atr,
                 dim, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_DIM_R(float* res_dat, int* res_atr,
                      float* src_dat, int* src_atr,
                      int* dim,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product_dim(res_dat, res_atr, src_dat, src_atr,
                 dim, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_DIM_D(double* res_dat, int* res_atr,
                      double* src_dat, int* src_atr,
                      int* dim,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product_dim(res_dat, res_atr, src_dat, src_atr,
                 dim, msk_dat, msk_atr, msk_present) ;
}

void AD_PRODUCT_DIM_Z(COMPLEX* res_dat, int* res_atr,
                      COMPLEX* src_dat, int* src_atr,
                      int* dim,
                      int* msk_dat, int* msk_atr, int* msk_present) {
  ad_product_dim(res_dat, res_atr, src_dat, src_atr,
                 dim, msk_dat, msk_atr, msk_present) ;
}


// Scalar-valued cumulative searches

// Maximum location

template<class T>
inline void ad_maxloc(int* pos_dat, int* pos_atr, T* src_dat, int* src_atr,
                      int* msk_dat, int* msk_atr, int* msk_present) {

  // NB: `pos_atr' argument is ignored (`pos' is known to be
  //     of global vector of length equal to the rank of `src').

  shpfDAD src(src_atr) ;
  T val ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MaxlocMsk<T> schedule(&src, &msk) ;
    schedule.execute(&val, pos_dat, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    Maxloc<T> schedule(&src) ;
    schedule.execute(&val, pos_dat, src_dat + src.offset) ;
  }
  for(int i = 0 ; i < src.rnk() ; i++)
    pos_dat [i]++ ;  // Fortran subscripts start at 1.
}

void AD_MAXLOC_I(int* res_dat, int* res_atr,
                 int* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}

void AD_MAXLOC_R(int* res_dat,   int* res_atr,
                 float* src_dat, int* src_atr,
                 int* msk_dat,   int* msk_atr, int* msk_present) {
  ad_maxloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}

void AD_MAXLOC_D(int* res_dat,    int* res_atr,
                 double* src_dat, int* src_atr,
                 int* msk_dat,    int* msk_atr, int* msk_present) {
  ad_maxloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}


// Minimum location

template<class T>
inline void ad_minloc(int* pos_dat, int* pos_atr, T* src_dat, int* src_atr,
                      int* msk_dat, int* msk_atr, int* msk_present) {

  // NB: `pos_atr' argument is ignored (`pos' is known to be
  //     of global vector of length equal to the rank of `src').

  shpfDAD src(src_atr) ;
  T val ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MinlocMsk<T> schedule(&src, &msk) ;
    schedule.execute(&val, pos_dat, src_dat + src.offset,
                     msk_dat + msk.offset) ;
  }
  else {
    Minloc<T> schedule(&src) ;
    schedule.execute(&val, pos_dat, src_dat + src.offset) ;
  }
  for(int i = 0 ; i < src.rnk() ; i++)
    pos_dat [i]++ ;  // Fortran subscripts start at 1.
}

void AD_MINLOC_I(int* res_dat, int* res_atr,
                 int* src_dat, int* src_atr,
                 int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}

void AD_MINLOC_R(int* res_dat,   int* res_atr,
                 float* src_dat, int* src_atr,
                 int* msk_dat,   int* msk_atr, int* msk_present) {
  ad_minloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}

void AD_MINLOC_D(int* res_dat,    int* res_atr,
                 double* src_dat, int* src_atr,
                 int* msk_dat,    int* msk_atr, int* msk_present) {
  ad_minloc(res_dat, res_atr, src_dat, src_atr,
            msk_dat, msk_atr, msk_present) ;
}


// Array-valued cumulative searches

// Maximum location

template<class T>
inline void ad_maxloc_dim(int* pos_dat, int* pos_atr, T* src_dat, int* src_atr,
                          int* dim,
                          int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD pos(pos_atr) ;
  shpfDAD src(src_atr) ;
  T* val_dat = new T [pos.size] ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MaxlocDimMsk<T> schedule(&pos, &pos, &src, *dim - 1, &msk, 1) ;
    schedule.execute(val_dat + pos.offset, pos_dat + pos.offset,
                     src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    MaxlocDim<T> schedule(&pos, &pos, &src, *dim - 1, 1) ;
    schedule.execute(val_dat + pos.offset, pos_dat + pos.offset,
                     src_dat + src.offset) ;
  }
  delete [] val_dat ;
}

void AD_MAXLOC_DIM_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  
void AD_MAXLOC_DIM_R(int* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  
void AD_MAXLOC_DIM_D(int* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_maxloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  

// Minimum location

template<class T>
inline void ad_minloc_dim(int* pos_dat, int* pos_atr, T* src_dat, int* src_atr,
                          int* dim,
                          int* msk_dat, int* msk_atr, int* msk_present) {
  shpfDAD pos(pos_atr) ;
  shpfDAD src(src_atr) ;
  T* val_dat = new T [pos.size] ;
  if(*msk_present) {
    shpfDAD msk(msk_atr) ;
    MinlocDimMsk<T> schedule(&pos, &pos, &src, *dim - 1, &msk, 1) ;
    schedule.execute(val_dat + pos.offset, pos_dat + pos.offset,
                     src_dat + src.offset, msk_dat + msk.offset) ;
  }
  else {
    MinlocDim<T> schedule(&pos, &pos, &src, *dim - 1, 1) ;
    schedule.execute(val_dat + pos.offset, pos_dat + pos.offset,
                     src_dat + src.offset) ;
  }
  delete [] val_dat ;
}

void AD_MINLOC_DIM_I(int* res_dat, int* res_atr,
                     int* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  
void AD_MINLOC_DIM_R(int* res_dat, int* res_atr,
                     float* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  
void AD_MINLOC_DIM_D(int* res_dat, int* res_atr,
                     double* src_dat, int* src_atr,
                     int* dim,
                     int* msk_dat, int* msk_atr, int* msk_present) {
  ad_minloc_dim(res_dat, res_atr, src_dat, src_atr,
                dim, msk_dat, msk_atr, msk_present) ;
}
  

// --------------------------------------------------------------------
// Things to do
//
//   Should do accumulations in double precision.
//
// 10) Modify kernel operations to allow generalized definition of
//     ``value-alignment''.
//
//   Gradually lose the old platform-specific versions of Adlib initialization.
//   Hopefully the new method works on most platforms, by now.

