
// Test `VecScatter', `VecScatterMsk' schedules

//   !hpf$ processors q(2, 3)
// 
//   !hpf$ template s(4, 10)
//   !hpf$ distribute s(block, block) onto q
// 
//   integer b(5, 5)
//   !hpf$ align b(*, j) with s(1, 8 - j)
// 
//   integer u(5), v(5)
//   !hpf$ align u(i) with b(i, *)
//   !hpf$ align v(j) with b(*, j)
// 
//   !hpf$ processors p(2, 3)
// 
//   !hpf$ template t(6, 3)
//   !hpf$ distribute t(cyclic, block) onto p
// 
//   integer a(6, 6)
// 
//   !hpf$ align a(i, *) with t(i, *)
// 
//   integer i, j
// 
//   do i = 1, 5
//     do j = 1, 5
//       b(i, j) = 10 * (i - 1) + (j - 1)
//     enddo
//   enddo
// 
//   do i = 1, 5
//     u(i) = 1 + mod((i - 1) * 11, 6)
//   enddo
// 
//   do j = 1, 5
//     v(j) = 1 + mod((j - 1) + 3, 6)
//   enddo
// 
//   do i = 1, 6
//     do j = 1, 6
//       a(i, j) = 0.0
//     enddo
//   enddo
// 
//   a(u, v) = b
// 
//   do i = 1, 6
//     do j = 1, 6
//       print "(2i3, i10)", i - 1, j - 1, a(i, j)
//     enddo
//   enddo
// 
// end


#include <stdio.h>
#include "ad++.h"


int main(int argc, char* argv[]) {
  AdlibInit(argc, argv) ;

  Procs2 qq(2, 3) ;

  //Range r(5) ;
  CollapsedRange r(5) ;
  Range s = BlockRange(10, qq.dim(1)).subrng(5, 6, -1) ;
  BlockRange t(4, qq.dim(0)) ;

  Group q(qq / t(0)) ;

  Array2<int> b(r, s, q) ;
  Array1<int> u(r, q) ;
  Array1<int> v(s, q) ;
  Array2<int> msk(r, s, q) ;

  Procs2 p(3, 2) ;

  CyclicRange x(6, p.dim(0)) ;
  //Range y(6) ;
  CollapsedRange y(6) ;
  Array2<int> a(x, y, p) ;

  ON(q) {
    Index ir(r), is(s) ;
    OVERALL(ir) {
      u(ir) = (ir * 11) % 6 ;

      OVERALL(is) {
	b(ir, is)   = 10 * ir + is ;
        msk(ir, is) = (ir + is) % 2 ;
      } ALLOVER(is) ;
    } ALLOVER(ir) ;

    OVERALL(is) {
      v(is) = (is + 3) % 6 ;
    } ALLOVER(is) ;
  } NO(q) ;

  ON(p) {
    Index ix(x), iy(y) ;
    OVERALL(ix) {
      OVERALL(iy) {
	a(ix, iy) = 0 ;
      } ALLOVER(iy) ;
    } ALLOVER(ix) ;
  } NO(p) ;

  vecScatter(b, a, &u, &v) ;

  gprintf("After vecScatter...\n") ;
  aprintf("%3R0%3R1%10d\n", a) ;

  ON(p) {
    Index ix(x), iy(y) ;
    OVERALL(ix) {
      OVERALL(iy) {
	a(ix, iy) = 0 ;
      } ALLOVER(iy) ;
    } ALLOVER(ix) ;
  } NO(p) ;

  vecScatter(b, a, &u, &v, msk) ;

  gprintf("After vecScatter with mask...\n") ;
  aprintf("%3R0%3R1%10d\n", a) ;

  AdlibFinalize() ;
}


