
// Test communication schedules for some large arrays.

#include "ad++.h"

#define N 800

int main(int argc, char* argv[]) {
  AdlibInit(argc, argv) ;

  Procs2 p(2, 2) ;
  ON(p) {
    BlockRange x(N, p.dim(0)) ;
    BlockRange y(N, p.dim(1)) ;

    Array2<int> a(x, y), b(x, y) ;

    Index i(x), j(y) ;
    
    OVERALL(i) {
      OVERALL(j) {
        a(i, j) = i * N + j ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;

    transpose(b, a) ;  // Uses `Remap' schedule internally.

    // (Recycle `a' as the comparision mask, to save space.)

    OVERALL(i) {
      OVERALL(j) {
        a(i, j) = b(i, j) == j * N + i ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;

    if(!all(a))
      gprintf("An error has occurred\n") ;
    else {

      // Double-check with `any', in case `all' is broken.

      OVERALL(i) {
	OVERALL(j) {
	  a(i, j) = b(i, j) != j * N + i ;
	} ALLOVER(j) ;
      } ALLOVER(i) ;

      if(any(a))
	gprintf("An error has occurred\n") ;
      else
	gprintf("Checked OK\n") ;
    }

    // Print a sample of the transposed matrix (in case `any' and
    // `all' are both broken).

    const int bx = N / 4, by = (3 * N) / 4 ;
    aprintf("%8i%8N", b.sect(Trip(bx, bx + 7), Trip(by, by + 7))) ;

  } NO(p) ;

  AdlibFinalize() ;
}

