
//   integer n
//   parameter(n = 8)
// 
//   real u(n, n)
// 
//   !hpf$ distribute u (block, block) onto p
//   !hpf$ processors p (2, 2)
// 
//   integer i, j
//   integer iter
// 
//   do i = 1, n
//     do j = 1, n
//       if(i == 1 .or. i == n .or. j == 1 .or. j == n) then
//         u(i, j) = 1.0
//       else
//         u(i, j) = 0.0
//       endif
//     enddo
//   enddo
// 
//   do iter = 1, 5
//     u(2 : n - 1, 2 : n - 1) = &
//                  0.25 * (u(2 : n - 1, 1 : n - 2) + u(2 : n - 1, 3 : n) + &
//                          u(1 : n - 2, 2 : n - 1) + u(3 : n, 2 : n - 1))
//     print *, "iteration =", iter
//     do i = 2, n - 1
//       print "(10f7.3)", u(i, 2 : n - 1)
//     enddo
//   enddo
// 
// end


#include <stdio.h>
#include "ad++.h"
#include "adrm.h"


const int n = 8 ;

int main(int argc, char* argv[]) {
  AdlibInit(argc, argv) ;
  ARMCI_Init() ;

  Procs2 p(2, 2) ;

  ON(p) {
    BlockRange x(n, p.dim(0)) ;
    BlockRange y(n, p.dim(1)) ;

    ArrayRM2<float> u(x, y) ;

    Index i(x), j(y) ;
    OVERALL(i) {
      OVERALL(j) {
	if(i == 0 || i == n - 1 || j == 0 || j == n - 1)
	  u(i, j) = 1.0 ;
	else
	  u(i, j) = 0.0 ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;

    Range xc = x.subrng(n - 2, 1) ;
    Range yc = y.subrng(n - 2, 1) ;

    SectionRM2<float> uLC = u.sect(Trip(0, n - 3), Trip(1, n - 2)) ;
    SectionRM2<float> uRC = u.sect(Trip(2, n - 1), Trip(1, n - 2)) ;
    SectionRM2<float> uCL = u.sect(Trip(1, n - 2), Trip(0, n - 3)) ;
    SectionRM2<float> uCR = u.sect(Trip(1, n - 2), Trip(2, n - 1)) ;

    Array2<float> tmp1(xc, yc), tmp2(xc, yc), tmp3(xc, yc), tmp4(xc, yc) ;

    for(int iter = 1 ; iter <= 5 ; iter++) {
      remapRM(tmp1, uLC) ;
      remapRM(tmp2, uRC) ;
      remapRM(tmp3, uCL) ;
      remapRM(tmp4, uCR) ;

      Index ic(xc), jc(yc) ;
      OVERALL(ic) {
	OVERALL(jc) {
	  u(ic, jc) = 0.25 * (tmp1(ic, jc) + tmp2(ic, jc) +
		              tmp3(ic, jc) + tmp4(ic, jc)) ;
	} ALLOVER(jc) ;
      } ALLOVER(ic) ;

      gprintf("iteration = %d\n", iter) ;
      aprintf("%7.3f%*N", u, n) ;
    }
  } NO(p) ;

  ARMCI_Finalize() ;
  AdlibFinalize() ;
}


